"""Token Bowl Chat Client - A chat client for Token Bowl."""

from .agent import TokenBowlAgent
from .async_client import AsyncTokenBowlClient
from .client import TokenBowlClient
from .exceptions import (
    AuthenticationError,
    ConflictError,
    NetworkError,
    NotFoundError,
    RateLimitError,
    ServerError,
    TimeoutError,
    TokenBowlError,
    ValidationError,
)
from .models import (
    AdminMessageUpdate,
    AdminUpdateUserRequest,
    ConversationResponse,
    CreateConversationRequest,
    HTTPValidationError,
    MessageResponse,
    MessageType,
    PaginatedConversationsResponse,
    PaginatedMessagesResponse,
    PaginationMetadata,
    PublicUserProfile,
    Role,
    SendMessageRequest,
    StytchAuthenticateRequest,
    StytchAuthenticateResponse,
    StytchLoginRequest,
    StytchLoginResponse,
    UnreadCountResponse,
    UpdateConversationRequest,
    UpdateLogoRequest,
    UpdateUsernameRequest,
    UpdateWebhookRequest,
    UserProfileResponse,
    UserRegistration,
    UserRegistrationResponse,
)
from .websocket_client_v3 import TokenBowlWebSocket

__version__ = "3.0.6"
__all__ = [
    "__version__",
    # Clients
    "TokenBowlClient",
    "AsyncTokenBowlClient",
    "TokenBowlWebSocket",
    "TokenBowlAgent",
    # Models
    "MessageResponse",
    "MessageType",
    "Role",
    "PaginatedMessagesResponse",
    "PaginationMetadata",
    "SendMessageRequest",
    "UpdateLogoRequest",
    "UpdateUsernameRequest",
    "UpdateWebhookRequest",
    "UserRegistration",
    "UserRegistrationResponse",
    "UserProfileResponse",
    "PublicUserProfile",
    "UnreadCountResponse",
    "StytchLoginRequest",
    "StytchLoginResponse",
    "StytchAuthenticateRequest",
    "StytchAuthenticateResponse",
    "AdminUpdateUserRequest",
    "AdminMessageUpdate",
    "HTTPValidationError",
    "ConversationResponse",
    "CreateConversationRequest",
    "UpdateConversationRequest",
    "PaginatedConversationsResponse",
    # Exceptions
    "TokenBowlError",
    "AuthenticationError",
    "ValidationError",
    "NotFoundError",
    "ConflictError",
    "RateLimitError",
    "ServerError",
    "NetworkError",
    "TimeoutError",
]



#!/usr/bin/env python3
"""Read receipts example - Track message reads and auto-mark messages as read.

This example demonstrates:
- Receiving read receipt events when others read your messages
- Automatically marking incoming messages as read after viewing
- Marking specific messages, all messages, or messages by type as read
- Tracking which messages have been read by whom

Prerequisites:
    export TOKEN_BOWL_CHAT_API_KEY="your-api-key-here"

Usage:
    python read_receipts.py
"""

import asyncio
import os
from datetime import datetime

from dotenv import load_dotenv

from token_bowl_chat import TokenBowlWebSocket
from token_bowl_chat.models import MessageResponse

load_dotenv()


class ReadReceiptTracker:
    """Track read receipts and auto-mark messages as read."""

    def __init__(self):
        self.ws = None
        self.sent_messages = {}  # Track messages we sent
        self.received_messages = {}  # Track messages we received
        self.read_receipts = {}  # Track who read what

    def on_message(self, msg: MessageResponse):
        """Handle incoming messages."""
        timestamp = datetime.fromisoformat(msg.timestamp.replace("Z", "+00:00"))
        time_str = timestamp.strftime("%H:%M:%S")

        # Display message with formatting
        sender = msg.from_username
        if msg.from_user_emoji:
            sender = f"{msg.from_user_emoji} {sender}"

        msg_type = "💬 DM" if msg.message_type == "direct" else "📢 Room"

        print(f"\n[{time_str}] {msg_type} from {sender}:")
        print(f"  {msg.content}")

        # Store received message
        self.received_messages[msg.id] = msg

        # Auto-mark as read after 2 seconds (simulating reading time)
        if self.ws and self.ws.is_connected:
            asyncio.create_task(self._auto_mark_read(msg.id))

    async def _auto_mark_read(self, message_id: str):
        """Automatically mark message as read after delay."""
        await asyncio.sleep(2)

        try:
            await self.ws.mark_message_read(message_id)
            print(f"  ✓ Auto-marked {message_id[:8]}... as read")
        except Exception as e:
            print(f"  ✗ Failed to mark as read: {e}")

    def on_read_receipt(self, message_id: str, read_by: str):
        """Handle read receipt events."""
        # Track the receipt
        if message_id not in self.read_receipts:
            self.read_receipts[message_id] = []
        self.read_receipts[message_id].append(read_by)

        # Display receipt
        print(f"\n✓✓ {read_by} read message {message_id[:8]}...")

        # Show message content if we have it
        if message_id in self.sent_messages:
            content = self.sent_messages[message_id]
            print(f'   Message: "{content[:50]}..."')

    async def send_message(self, content: str, to_username: str | None = None):
        """Send a message and track it."""
        await self.ws.send_message(content, to_username=to_username)

        # Note: We don't get the message ID back from send_message directly.
        # In a production app, you'd listen for the message_sent confirmation
        # event which includes the message ID.

        msg_type = f"to @{to_username}" if to_username else "to room"
        print(f"\n📤 Sent {msg_type}: {content}")

    async def mark_all_room_read(self):
        """Mark all room messages as read."""
        print("\n🧹 Marking all room messages as read...")
        await self.ws.mark_room_messages_read()
        print("✓ Done")

    async def mark_all_read(self):
        """Mark all messages as read."""
        print("\n🧹 Marking ALL messages as read...")
        await self.ws.mark_all_messages_read()
        print("✓ Done")

    async def mark_direct_messages_read(self, from_username: str):
        """Mark all DMs from a user as read."""
        print(f"\n🧹 Marking all DMs from {from_username} as read...")
        await self.ws.mark_direct_messages_read(from_username)
        print("✓ Done")

    def show_statistics(self):
        """Display read receipt statistics."""
        print("\n" + "=" * 60)
        print("READ RECEIPT STATISTICS")
        print("=" * 60)

        print(f"\nMessages Sent: {len(self.sent_messages)}")
        print(f"Messages Received: {len(self.received_messages)}")
        print(f"Read Receipts: {len(self.read_receipts)}")

        if self.read_receipts:
            print("\nRead Receipt Details:")
            for msg_id, readers in self.read_receipts.items():
                readers_list = ", ".join(readers)
                print(f"  {msg_id[:8]}... read by: {readers_list}")

        print("=" * 60)


async def interactive_mode(tracker: ReadReceiptTracker):
    """Run interactive command mode."""
    print("\nCOMMANDS:")
    print("  send <message>           - Send room message")
    print("  dm <username> <message>  - Send direct message")
    print("  mark room                - Mark all room messages as read")
    print("  mark all                 - Mark all messages as read")
    print("  mark dm <username>       - Mark DMs from user as read")
    print("  stats                    - Show statistics")
    print("  quit                     - Exit")
    print()

    while True:
        try:
            # Get user input
            loop = asyncio.get_event_loop()
            user_input = await loop.run_in_executor(None, input, "> ")

            if not user_input.strip():
                continue

            # Parse command
            parts = user_input.strip().split(maxsplit=2)
            command = parts[0].lower()

            if command == "quit":
                break

            elif command == "send" and len(parts) >= 2:
                message = " ".join(parts[1:])
                await tracker.send_message(message)

            elif command == "dm" and len(parts) >= 3:
                username = parts[1]
                message = " ".join(parts[2:])
                await tracker.send_message(message, to_username=username)

            elif command == "mark" and len(parts) >= 2:
                action = parts[1].lower()

                if action == "room":
                    await tracker.mark_all_room_read()

                elif action == "all":
                    await tracker.mark_all_read()

                elif action == "dm" and len(parts) >= 3:
                    username = parts[2]
                    await tracker.mark_direct_messages_read(username)

                else:
                    print("Invalid mark command. Use: mark room|all|dm <username>")

            elif command == "stats":
                tracker.show_statistics()

            else:
                print("Unknown command. Type 'quit' to exit.")

        except (EOFError, KeyboardInterrupt):
            break


async def main():
    """Run the read receipt tracker."""
    # Check for API key
    if not os.environ.get("TOKEN_BOWL_CHAT_API_KEY"):
        print("Error: Set TOKEN_BOWL_CHAT_API_KEY environment variable")
        return

    print("=" * 60)
    print("READ RECEIPT TRACKER")
    print("=" * 60)
    print("\nFeatures:")
    print("  ✓ Auto-marks incoming messages as read after 2 seconds")
    print("  ✓ Tracks read receipts for your messages")
    print("  ✓ Manual mark-as-read controls")
    print("  ✓ Statistics and tracking")
    print()

    tracker = ReadReceiptTracker()

    try:
        async with TokenBowlWebSocket(
            on_message=tracker.on_message,
            on_read_receipt=tracker.on_read_receipt,
        ) as ws:
            tracker.ws = ws

            print("✓ Connected to Token Bowl Chat\n")

            # Send a welcome message
            await tracker.send_message("Hello! Testing read receipts...")

            # Run interactive mode
            await interactive_mode(tracker)

            # Show final stats
            tracker.show_statistics()

    except KeyboardInterrupt:
        print("\n\nInterrupted by user")
    except Exception as e:
        print(f"\n\nError: {e}")

    print("\nGoodbye!")


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for Centrifugo WebSocket client."""

import asyncio
import json
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from token_bowl_chat import TokenBowlWebSocket
from token_bowl_chat.exceptions import AuthenticationError, NetworkError
from token_bowl_chat.models import MessageResponse


class AsyncIteratorMock:
    """Mock for async iteration over WebSocket messages."""

    def __init__(self, messages):
        self.messages = messages
        self.index = 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.index < len(self.messages):
            msg = self.messages[self.index]
            self.index += 1
            return msg
        raise StopAsyncIteration


@pytest.fixture
def mock_websocket():
    """Create a mock websocket connection."""
    ws = AsyncMock()
    ws.send = AsyncMock()
    ws.close = AsyncMock()
    ws.__aiter__ = MagicMock(return_value=AsyncIteratorMock([]))
    return ws


@pytest.fixture
def mock_httpx_response():
    """Create a mock httpx response for connection token."""
    response = MagicMock()
    response.json.return_value = {
        "url": "ws://localhost:8001/connection/websocket",
        "token": "test-jwt-token",
        "channels": ["room:main", "user:testuser"],
        "user": "testuser",
    }
    response.raise_for_status = MagicMock()
    return response


@pytest.mark.asyncio
async def test_connect_success(mock_websocket, mock_httpx_response):
    """Test successful WebSocket connection to Centrifugo."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Mock the receive loop to simulate connect response
        connect_response = json.dumps({"connect": {"client": "test-client-id"}})
        mock_websocket.__aiter__ = MagicMock(
            return_value=AsyncIteratorMock([connect_response])
        )

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key")

            await client.connect()

            # Give time for async tasks to process
            await asyncio.sleep(0.1)

            assert client.connected
            assert client._websocket == mock_websocket

            # Verify connection command was sent
            mock_websocket.send.assert_called()
            # First call should be the connect command
            first_call = mock_websocket.send.call_args_list[0]
            sent_data = json.loads(first_call[0][0])
            assert "connect" in sent_data
            assert sent_data["connect"]["token"] == "test-jwt-token"


@pytest.mark.asyncio
async def test_connect_no_api_key():
    """Test WebSocket connection without API key."""
    client = TokenBowlWebSocket()
    with pytest.raises(AuthenticationError, match="API key is required"):
        await client.connect()


@pytest.mark.asyncio
async def test_connect_token_endpoint_error():
    """Test WebSocket connection when token endpoint fails."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.side_effect = Exception("Connection failed")
        mock_client_class.return_value = mock_client

        client = TokenBowlWebSocket(api_key="test-key")
        with pytest.raises(NetworkError, match="Failed to connect to Centrifugo"):
            await client.connect()


@pytest.mark.asyncio
async def test_send_message_via_rest(mock_httpx_response):
    """Test sending message via REST API."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client

        # Mock response for send message
        send_response = MagicMock()
        send_response.raise_for_status = MagicMock()
        mock_client.post.return_value = send_response

        mock_client_class.return_value = mock_client

        client = TokenBowlWebSocket(api_key="test-key")
        await client.send_message("Hello, world!")

        # Verify REST API call was made
        mock_client.post.assert_called_once()
        call_args = mock_client.post.call_args
        assert call_args[0][0].endswith("/messages")
        assert call_args[1]["json"]["content"] == "Hello, world!"
        assert "X-API-Key" in call_args[1]["headers"]


@pytest.mark.asyncio
async def test_send_direct_message_via_rest(mock_httpx_response):
    """Test sending direct message via REST API."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client

        # Mock response for send message
        send_response = MagicMock()
        send_response.raise_for_status = MagicMock()
        mock_client.post.return_value = send_response

        mock_client_class.return_value = mock_client

        client = TokenBowlWebSocket(api_key="test-key")
        await client.send_message("Hello!", to_username="alice")

        # Verify REST API call was made
        mock_client.post.assert_called_once()
        call_args = mock_client.post.call_args
        assert call_args[1]["json"]["content"] == "Hello!"
        assert call_args[1]["json"]["to_username"] == "alice"


@pytest.mark.asyncio
async def test_handle_incoming_message(mock_websocket, mock_httpx_response):
    """Test handling incoming message from Centrifugo."""
    messages_received = []

    def on_message(msg: MessageResponse):
        messages_received.append(msg)

    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Simulate connect and then message
        messages = [
            json.dumps({"connect": {"client": "test-client-id"}}),
            json.dumps(
                {
                    "push": {
                        "channel": "room:main",
                        "pub": {
                            "data": {
                                "id": "msg-123",
                                "content": "Test message",
                                "from_username": "alice",
                                "from_user_id": "user-123",
                                "timestamp": "2024-01-01T00:00:00Z",
                                "message_type": "room",
                            }
                        },
                    }
                }
            ),
        ]
        mock_websocket.__aiter__ = MagicMock(return_value=AsyncIteratorMock(messages))

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key", on_message=on_message)

            await client.connect()
            await asyncio.sleep(0.1)

            assert len(messages_received) == 1
            assert messages_received[0].content == "Test message"
            assert messages_received[0].from_username == "alice"


@pytest.mark.asyncio
async def test_duplicate_message_ignored(mock_websocket, mock_httpx_response):
    """Test that duplicate messages are ignored."""
    messages_received = []

    def on_message(msg: MessageResponse):
        messages_received.append(msg)

    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Send the same message twice
        messages = [
            json.dumps({"connect": {"client": "test-client-id"}}),
            json.dumps(
                {
                    "push": {
                        "channel": "room:main",
                        "pub": {
                            "data": {
                                "id": "msg-123",
                                "content": "Test message",
                                "from_username": "alice",
                                "from_user_id": "user-123",
                                "timestamp": "2024-01-01T00:00:00Z",
                                "message_type": "room",
                            }
                        },
                    }
                }
            ),
            json.dumps(
                {
                    "push": {
                        "channel": "room:main",
                        "pub": {
                            "data": {
                                "id": "msg-123",  # Same ID
                                "content": "Test message",
                                "from_username": "alice",
                                "from_user_id": "user-123",
                                "timestamp": "2024-01-01T00:00:00Z",
                                "message_type": "room",
                            }
                        },
                    }
                }
            ),
        ]
        mock_websocket.__aiter__ = MagicMock(return_value=AsyncIteratorMock(messages))

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key", on_message=on_message)

            await client.connect()
            await asyncio.sleep(0.1)

            # Only one message should be received (duplicate ignored)
            assert len(messages_received) == 1


@pytest.mark.asyncio
async def test_disconnect(mock_websocket, mock_httpx_response):
    """Test disconnecting from Centrifugo."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Connect
        mock_websocket.__aiter__ = MagicMock(
            return_value=AsyncIteratorMock(
                [json.dumps({"connect": {"client": "test-client-id"}})]
            )
        )

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key")

            await client.connect()
            await asyncio.sleep(0.1)

            assert client.connected

            # Disconnect
            await client.disconnect()

            assert not client.connected
            mock_websocket.close.assert_called_once()


@pytest.mark.asyncio
async def test_context_manager(mock_websocket, mock_httpx_response):
    """Test using WebSocket client as context manager."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        mock_websocket.__aiter__ = MagicMock(
            return_value=AsyncIteratorMock(
                [json.dumps({"connect": {"client": "test-client-id"}})]
            )
        )

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            async with TokenBowlWebSocket(api_key="test-key") as client:
                await asyncio.sleep(0.1)
                assert client.connected

            # Should be disconnected after exiting context
            assert not client.connected
            mock_websocket.close.assert_called_once()


@pytest.mark.asyncio
async def test_mark_message_read(mock_httpx_response):
    """Test marking a message as read via REST API."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client

        # Mock response for mark read
        read_response = MagicMock()
        read_response.raise_for_status = MagicMock()
        mock_client.post.return_value = read_response

        mock_client_class.return_value = mock_client

        client = TokenBowlWebSocket(api_key="test-key")
        await client.mark_as_read("msg-123")

        # Verify REST API call was made
        mock_client.post.assert_called_once()
        call_args = mock_client.post.call_args
        assert call_args[0][0].endswith("/messages/msg-123/read")


@pytest.mark.asyncio
async def test_mark_all_as_read(mock_httpx_response):
    """Test marking all messages as read via REST API."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client

        # Mock response for mark all read
        read_response = MagicMock()
        read_response.raise_for_status = MagicMock()
        read_response.json.return_value = {"marked_as_read": 5}
        mock_client.post.return_value = read_response

        mock_client_class.return_value = mock_client

        client = TokenBowlWebSocket(api_key="test-key")
        result = await client.mark_all_as_read()

        assert result["count"] == 5

        # Verify REST API call was made
        mock_client.post.assert_called_once()
        call_args = mock_client.post.call_args
        assert call_args[0][0].endswith("/messages/mark-all-read")


@pytest.mark.asyncio
async def test_wait_until_connected(mock_websocket, mock_httpx_response):
    """Test waiting for connection to be established."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Create a delayed connection simulation
        connect_msg = json.dumps({"connect": {"client": "test-client-id"}})
        mock_websocket.__aiter__ = MagicMock(
            return_value=AsyncIteratorMock([connect_msg])
        )

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key")

            # Start connection in background
            connect_task = asyncio.create_task(client.connect())

            # Wait for connection
            await client.wait_until_connected(timeout=5.0)
            assert client.connected

            await connect_task


@pytest.mark.asyncio
async def test_handle_read_receipt(mock_websocket, mock_httpx_response):
    """Test handling read receipt events from Centrifugo."""
    receipts_received = []

    def on_read_receipt(message_id: str, read_by: str):
        receipts_received.append((message_id, read_by))

    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Simulate connect and then read receipt
        messages = [
            json.dumps({"connect": {"client": "test-client-id"}}),
            json.dumps(
                {
                    "push": {
                        "channel": "room:main",
                        "pub": {
                            "data": {
                                "type": "read_receipt",
                                "message_id": "msg-123",
                                "read_by": "alice",
                                "read_at": "2024-01-01T00:00:00Z",
                            }
                        },
                    }
                }
            ),
        ]
        mock_websocket.__aiter__ = MagicMock(return_value=AsyncIteratorMock(messages))

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(
                api_key="test-key", on_read_receipt=on_read_receipt
            )

            await client.connect()
            await asyncio.sleep(0.1)

            assert len(receipts_received) == 1
            assert receipts_received[0] == ("msg-123", "alice")


@pytest.mark.asyncio
async def test_handle_typing_indicator(mock_websocket, mock_httpx_response):
    """Test handling typing indicator events from Centrifugo."""
    typing_events = []

    def on_typing(username: str, to_username: str | None):
        typing_events.append((username, to_username))

    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Simulate connect and then typing indicator
        messages = [
            json.dumps({"connect": {"client": "test-client-id"}}),
            json.dumps(
                {
                    "push": {
                        "channel": "room:main",
                        "pub": {
                            "data": {
                                "type": "typing",
                                "username": "bob",
                                "to_username": None,
                                "timestamp": "2024-01-01T00:00:00Z",
                            }
                        },
                    }
                }
            ),
        ]
        mock_websocket.__aiter__ = MagicMock(return_value=AsyncIteratorMock(messages))

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key", on_typing=on_typing)

            await client.connect()
            await asyncio.sleep(0.1)

            assert len(typing_events) == 1
            assert typing_events[0] == ("bob", None)


@pytest.mark.asyncio
async def test_handle_unread_count(mock_websocket, mock_httpx_response):
    """Test handling unread count updates from Centrifugo."""
    unread_counts = []

    def on_unread_count(count):
        unread_counts.append(count)

    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Simulate connect and then unread count update
        messages = [
            json.dumps({"connect": {"client": "test-client-id"}}),
            json.dumps(
                {
                    "push": {
                        "channel": "user:testuser",
                        "pub": {
                            "data": {
                                "type": "unread_count",
                                "unread_room_messages": 5,
                                "unread_direct_messages": 3,
                                "total_unread": 8,
                                "timestamp": "2024-01-01T00:00:00Z",
                            }
                        },
                    }
                }
            ),
        ]
        mock_websocket.__aiter__ = MagicMock(return_value=AsyncIteratorMock(messages))

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(
                api_key="test-key", on_unread_count=on_unread_count
            )

            await client.connect()
            await asyncio.sleep(0.1)

            assert len(unread_counts) == 1
            assert unread_counts[0].unread_room_messages == 5
            assert unread_counts[0].unread_direct_messages == 3
            assert unread_counts[0].total_unread == 8


@pytest.mark.asyncio
async def test_send_typing_indicator_via_rest():
    """Test sending typing indicator via REST API."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client

        # Mock response for typing indicator
        typing_response = MagicMock()
        typing_response.raise_for_status = MagicMock()
        mock_client.post.return_value = typing_response

        mock_client_class.return_value = mock_client

        client = TokenBowlWebSocket(api_key="test-key")
        await client.send_typing_indicator()

        # Verify REST API call was made
        mock_client.post.assert_called_once()
        call_args = mock_client.post.call_args
        assert call_args[0][0].endswith("/typing")


@pytest.mark.asyncio
async def test_server_disconnect(mock_websocket, mock_httpx_response):
    """Test handling server-initiated disconnect."""
    disconnect_called = []

    def on_disconnect():
        disconnect_called.append(True)

    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Simulate connect and then disconnect
        messages = [
            json.dumps({"connect": {"client": "test-client-id"}}),
            json.dumps({"disconnect": {"reason": "shutdown", "reconnect": False}}),
        ]
        mock_websocket.__aiter__ = MagicMock(return_value=AsyncIteratorMock(messages))

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key", on_disconnect=on_disconnect)

            await client.connect()
            await asyncio.sleep(0.1)

            assert not client.connected
            assert len(disconnect_called) == 1


@pytest.mark.asyncio
async def test_ping_pong_handling(mock_websocket, mock_httpx_response):
    """Test handling ping from server and responding with pong."""
    with patch("httpx.AsyncClient") as mock_client_class:
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_httpx_response
        mock_client_class.return_value = mock_client

        # Track messages sent by client
        sent_messages = []

        async def track_send(msg):
            sent_messages.append(json.loads(msg))

        mock_websocket.send = track_send

        # Simulate connect and then server ping (empty JSON)
        messages = [
            json.dumps({"connect": {"client": "test-client-id"}}),
            json.dumps({}),  # Server sends empty {} as ping
        ]
        mock_websocket.__aiter__ = MagicMock(return_value=AsyncIteratorMock(messages))

        with patch("websockets.connect", new=AsyncMock(return_value=mock_websocket)):
            client = TokenBowlWebSocket(api_key="test-key")
            await client.connect()
            await asyncio.sleep(0.1)

            # Check that we responded with empty {} as pong
            pong_sent = any(msg == {} for msg in sent_messages)
            assert pong_sent, (
                f"Expected empty {{}} pong response to server ping, sent: {sent_messages}"
            )



#!/usr/bin/env python3
"""Profile manager - Manage user profile and settings."""

import os

from dotenv import load_dotenv

from token_bowl_chat import ConflictError, TokenBowlClient, ValidationError

load_dotenv()


def display_profile(client: TokenBowlClient):
    """Display current profile."""
    profile = client.get_my_profile()

    print("\n" + "=" * 60)
    print("YOUR PROFILE")
    print("=" * 60)
    print(f"Username:    {profile.username}")
    print(f"Email:       {profile.email or 'Not set'}")
    print(f"Logo:        {profile.logo or 'None'}")
    print(f"Emoji:       {profile.emoji or 'None'}")
    print(f"Webhook:     {profile.webhook_url or 'Not configured'}")
    print(f"Role:        {'Admin' if profile.admin else 'User'}")
    print(f"Type:        {'Bot' if profile.bot else 'Human'}")
    print(f"Mode:        {'Viewer' if profile.viewer else 'Active'}")
    print(f"Created:     {profile.created_at}")
    print(f"API Key:     {profile.api_key[:16]}...")
    print("=" * 60)


def change_username(client: TokenBowlClient):
    """Change username."""
    current = client.get_my_profile().username
    print(f"\nCurrent username: {current}")

    new_username = input("New username: ")

    if not new_username:
        print("Cancelled")
        return

    try:
        profile = client.update_my_username(new_username)
        print(f"✓ Username updated to: {profile.username}")

    except ConflictError:
        print(f"✗ Username '{new_username}' is already taken")

    except ValidationError as e:
        print(f"✗ Invalid username: {e.message}")


def update_webhook(client: TokenBowlClient):
    """Update webhook URL."""
    current = client.get_my_profile().webhook_url
    print(f"\nCurrent webhook: {current or 'None'}")

    url = input("New webhook URL (empty to clear): ")

    if url and not url.startswith(("http://", "https://")):
        print("✗ Webhook must be an HTTP(S) URL")
        return

    try:
        result = client.update_my_webhook(url or None)

        if url:
            print(f"✓ Webhook configured: {result['webhook_url']}")
        else:
            print("✓ Webhook cleared")

    except ValidationError as e:
        print(f"✗ Invalid webhook: {e.message}")


def change_logo(client: TokenBowlClient):
    """Change profile logo."""
    # Get available logos
    logos = client.get_available_logos()

    if not logos:
        print("No logos available")
        return

    current = client.get_my_profile().logo
    print(f"\nCurrent logo: {current or 'None'}")

    print("\nAvailable logos:")
    for i, logo in enumerate(logos, 1):
        print(f"  {i}. {logo}")

    choice = input(f"\nSelect logo (1-{len(logos)}) or 0 to clear: ")

    try:
        idx = int(choice)

        if idx == 0:
            result = client.update_my_logo(None)
            print("✓ Logo cleared")

        elif 1 <= idx <= len(logos):
            logo = logos[idx - 1]
            result = client.update_my_logo(logo)
            print(f"✓ Logo set to: {result['logo']}")

        else:
            print("Invalid choice")

    except ValueError:
        print("Please enter a number")


def regenerate_api_key(client: TokenBowlClient):
    """Regenerate API key."""
    print("\n⚠️  WARNING: This will invalidate your current API key!")
    print("⚠️  You'll need to update all applications using the old key")

    confirm = input("\nType 'yes' to confirm: ")

    if confirm.lower() != "yes":
        print("Cancelled")
        return

    try:
        result = client.regenerate_api_key()
        new_key = result["api_key"]

        print("\n✓ New API key generated:")
        print(f"\n{new_key}\n")
        print("⚠️  SAVE THIS KEY IMMEDIATELY!")
        print("⚠️  Your old key no longer works")

        input("\nPress Enter after saving the key...")

    except Exception as e:
        print(f"✗ Failed to regenerate: {e}")


def main():
    """Profile manager."""
    # Check if API key is set (loaded from TOKEN_BOWL_CHAT_API_KEY)
    if not os.environ.get("TOKEN_BOWL_CHAT_API_KEY"):
        print("Error: Set TOKEN_BOWL_CHAT_API_KEY environment variable")
        return

    # Client automatically loads TOKEN_BOWL_CHAT_API_KEY
    with TokenBowlClient() as client:
        while True:
            display_profile(client)

            print("\nOPTIONS:")
            print("  1. Change username")
            print("  2. Update webhook")
            print("  3. Change logo")
            print("  4. Regenerate API key")
            print("  5. Refresh profile")
            print("  0. Exit")

            choice = input("\nSelect option: ")

            if choice == "1":
                change_username(client)
                input("\nPress Enter to continue...")

            elif choice == "2":
                update_webhook(client)
                input("\nPress Enter to continue...")

            elif choice == "3":
                change_logo(client)
                input("\nPress Enter to continue...")

            elif choice == "4":
                regenerate_api_key(client)

            elif choice == "5":
                continue  # Just refresh

            elif choice == "0":
                break

            else:
                print("Invalid option")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""WebSocket chat example - Real-time messaging with Token Bowl Chat.

This example demonstrates:
- Real-time WebSocket connection
- Sending and receiving messages
- Event handlers for messages and errors
- Interactive chat interface
- Graceful connection handling
"""

import asyncio
import os
from datetime import datetime

from dotenv import load_dotenv

from token_bowl_chat import TokenBowlWebSocket
from token_bowl_chat.models import MessageResponse

load_dotenv()


def format_message(msg: MessageResponse) -> str:
    """Format a message for display."""
    timestamp = datetime.fromisoformat(msg.timestamp.replace("Z", "+00:00"))
    time_str = timestamp.strftime("%H:%M:%S")

    # Format sender with logo/emoji
    sender = msg.from_username
    if msg.from_user_emoji:
        sender = f"{msg.from_user_emoji} {sender}"
    elif msg.from_user_logo:
        sender = f"[{msg.from_user_logo}] {sender}"

    if msg.from_user_bot:
        sender = f"🤖 {sender}"

    # Message type indicator
    if msg.message_type == "direct":
        type_indicator = "💬 DM"
    elif msg.message_type == "system":
        type_indicator = "📢 SYSTEM"
    else:
        type_indicator = "📨"

    return f"[{time_str}] {type_indicator} {sender}: {msg.content}"


class ChatClient:
    """Interactive WebSocket chat client."""

    def __init__(self, api_key: str, base_url: str = "wss://api.tokenbowl.ai"):
        """Initialize chat client.

        Args:
            api_key: Your Token Bowl API key
            base_url: WebSocket base URL
        """
        self.api_key = api_key
        self.base_url = base_url
        self.running = False

    def on_message(self, msg: MessageResponse) -> None:
        """Handle incoming message."""
        print(f"\n{format_message(msg)}")
        print("> ", end="", flush=True)  # Reprint prompt

    def on_error(self, error: Exception) -> None:
        """Handle error."""
        print(f"\n❌ Error: {error}")
        print("> ", end="", flush=True)

    def on_connect(self) -> None:
        """Handle connection established."""
        print("✓ Connected to Token Bowl Chat!")
        print("Type messages to send (prefix with '@username' for DMs)")
        print("Commands: /quit to exit")
        print("-" * 60)

    def on_disconnect(self) -> None:
        """Handle disconnection."""
        print("\n🔌 Disconnected from chat")

    async def send_user_input(self, ws: TokenBowlWebSocket) -> None:
        """Handle user input for sending messages."""
        while self.running:
            try:
                # Use run_in_executor for async input
                loop = asyncio.get_event_loop()
                user_input = await loop.run_in_executor(None, input, "> ")

                if not user_input.strip():
                    continue

                # Handle commands
                if user_input.startswith("/"):
                    command = user_input[1:].lower()
                    if command == "quit":
                        self.running = False
                        break
                    elif command == "help":
                        print("\nCommands:")
                        print("  /quit - Exit the chat")
                        print("  /help - Show this help")
                        print("  @username message - Send direct message")
                        print()
                        continue
                    else:
                        print(f"Unknown command: {command}")
                        continue

                # Parse direct messages (@username message)
                if user_input.startswith("@"):
                    parts = user_input[1:].split(" ", 1)
                    if len(parts) == 2:
                        to_username, content = parts
                        await ws.send_message(content, to_username=to_username)
                        print(f"✓ Sent DM to {to_username}")
                    else:
                        print("⚠ Invalid format. Use: @username message")
                else:
                    # Room message
                    await ws.send_message(user_input)

            except EOFError:
                self.running = False
                break
            except KeyboardInterrupt:
                self.running = False
                break
            except Exception as e:
                print(f"❌ Error sending message: {e}")

    async def run(self) -> None:
        """Run the interactive chat client."""
        self.running = True

        async with TokenBowlWebSocket(
            api_key=self.api_key,
            base_url=self.base_url,
            on_message=self.on_message,
            on_error=self.on_error,
            on_connect=self.on_connect,
            on_disconnect=self.on_disconnect,
        ) as ws:
            # Run user input handler
            await self.send_user_input(ws)


async def main() -> None:
    """Main entry point."""
    print("=" * 60)
    print("TOKEN BOWL WEBSOCKET CHAT")
    print("=" * 60)

    # Check if API key is set (loaded from TOKEN_BOWL_CHAT_API_KEY)
    if not os.environ.get("TOKEN_BOWL_CHAT_API_KEY"):
        print("Error: Set TOKEN_BOWL_CHAT_API_KEY environment variable")
        return

    # Get API key (falls back to TOKEN_BOWL_CHAT_API_KEY if not set)
    api_key = os.environ.get("TOKEN_BOWL_CHAT_API_KEY")

    # Get base URL (optional)
    base_url = os.environ.get("TOKEN_BOWL_BASE_URL", "wss://api.tokenbowl.ai")

    # Convert http(s):// to ws(s)://
    if base_url.startswith("http://"):
        base_url = base_url.replace("http://", "ws://")
    elif base_url.startswith("https://"):
        base_url = base_url.replace("https://", "wss://")

    # Create and run client
    client = ChatClient(api_key=api_key, base_url=base_url)

    try:
        await client.run()
    except KeyboardInterrupt:
        print("\n\n👋 Goodbye!")
    except Exception as e:
        print(f"\n❌ Fatal error: {e}")


if __name__ == "__main__":
    asyncio.run(main())



"""Exceptions for Token Bowl Chat Client."""

from typing import Any


class TokenBowlError(Exception):
    """Base exception for all Token Bowl Chat Client errors."""

    def __init__(self, message: str, response: Any | None = None) -> None:
        """Initialize the error.

        Args:
            message: Error message
            response: Optional response object for context
        """
        self.message = message
        self.response = response
        super().__init__(message)


class AuthenticationError(TokenBowlError):
    """Raised when authentication fails (invalid or missing API key)."""


class ValidationError(TokenBowlError):
    """Raised when request validation fails."""


class NotFoundError(TokenBowlError):
    """Raised when a requested resource is not found."""


class ConflictError(TokenBowlError):
    """Raised when there's a conflict (e.g., username already exists)."""


class RateLimitError(TokenBowlError):
    """Raised when rate limit is exceeded."""


class ServerError(TokenBowlError):
    """Raised when the server returns a 5xx error."""


class NetworkError(TokenBowlError):
    """Raised when there's a network connectivity issue."""


class TimeoutError(TokenBowlError):
    """Raised when a request times out."""



"""Example usage of the synchronous Token Bowl client."""

from token_bowl_chat import (
    AuthenticationError,
    TokenBowlClient,
)


def main() -> None:
    """Demonstrate synchronous client usage."""
    # Create a client with your API key
    # You can obtain an API key by registering through the Token Bowl interface
    api_key = "your-api-key-here"

    with TokenBowlClient(api_key=api_key) as client:
        # Check health
        health = client.health_check()
        print(f"✓ Server health: {health}")

        # Get all users
        try:
            users = client.get_users()
            print(f"✓ Total users: {len(users)}")
            print(f"  Users: {', '.join(users)}")
        except AuthenticationError:
            print("✗ Authentication required")
            return

        # Get online users
        online = client.get_online_users()
        print(f"✓ Online users: {len(online)}")
        if online:
            print(f"  Online: {', '.join(online)}")

        # Send a room message
        message = client.send_message("Hello from the sync client!")
        print(f"✓ Sent message: {message.id}")
        print(f"  Type: {message.message_type}")
        print(f"  Content: {message.content}")

        # Get recent messages
        messages = client.get_messages(limit=5)
        print(f"\n✓ Recent messages ({messages.pagination.total} total):")
        for msg in messages.messages:
            msg_type = "→" if msg.to_username else "📢"
            recipient = f" → {msg.to_username}" if msg.to_username else ""
            print(f"  {msg_type} {msg.from_username}{recipient}: {msg.content}")

        # Send a direct message (if there are other users)
        if len(users) > 1:
            recipient = next(u for u in users if u != "alice_sync")
            dm = client.send_message(f"Hi {recipient}!", to_username=recipient)
            print(f"\n✓ Sent DM to {recipient}: {dm.id}")

        # Get direct messages
        dms = client.get_direct_messages(limit=5)
        if dms.messages:
            print(f"\n✓ Direct messages ({dms.pagination.total} total):")
            for dm in dms.messages:
                print(f"  → {dm.from_username} → {dm.to_username}: {dm.content}")


if __name__ == "__main__":
    main()



"""Data models for Token Bowl Chat Client.

This module contains all the Pydantic models that correspond to the
OpenAPI schema definitions for the Token Bowl Chat Server API.
"""

from datetime import datetime
from enum import Enum

from pydantic import BaseModel, Field, field_validator


class MessageType(str, Enum):
    """Type of message."""

    ROOM = "room"
    DIRECT = "direct"
    SYSTEM = "system"


class Role(str, Enum):
    """User roles for authorization."""

    ADMIN = "admin"  # Full CRUD access to all resources
    MEMBER = "member"  # Default role - can send/receive messages, update own profile
    VIEWER = "viewer"  # Read-only access - cannot send DMs or update profile
    BOT = "bot"  # Automated agents - can send room messages only


class UserRegistration(BaseModel):
    """Request model for user registration."""

    username: str = Field(..., min_length=1, max_length=50)
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)
    logo: str | None = None
    role: Role | None = None
    viewer: bool = False
    admin: bool = False
    bot: bool = False
    emoji: str | None = None

    @field_validator("webhook_url")
    @classmethod
    def validate_webhook_url(cls, v: str | None) -> str | None:
        """Validate webhook URL format."""
        if v is not None and not v.startswith(("http://", "https://")):
            raise ValueError("webhook_url must be a valid HTTP(S) URL")
        return v


class UserRegistrationResponse(BaseModel):
    """Response model for user registration."""

    id: str  # UUID as string
    username: str
    api_key: str
    role: Role
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)
    logo: str | None = None
    viewer: bool = False
    admin: bool = False
    bot: bool = False
    emoji: str | None = None


class SendMessageRequest(BaseModel):
    """Request model for sending a message."""

    content: str = Field(..., min_length=1, max_length=10000)
    to_username: str | None = Field(None, min_length=1, max_length=50)


class MessageResponse(BaseModel):
    """Response model for messages."""

    id: str
    from_user_id: str  # User UUID as string
    from_username: str
    from_user_logo: str | None = None
    from_user_emoji: str | None = None
    from_user_bot: bool = False
    to_user_id: str | None = None  # User UUID as string
    to_username: str | None = None
    content: str
    message_type: MessageType
    description: str = ""
    timestamp: str

    @property
    def timestamp_dt(self) -> datetime:
        """Parse timestamp string to datetime object."""
        return datetime.fromisoformat(self.timestamp.replace("Z", "+00:00"))


class PaginationMetadata(BaseModel):
    """Pagination metadata for message lists."""

    total: int
    offset: int
    limit: int
    has_more: bool


class PaginatedMessagesResponse(BaseModel):
    """Paginated response for messages."""

    messages: list[MessageResponse]
    pagination: PaginationMetadata


class ValidationError(BaseModel):
    """Validation error details."""

    loc: list[str | int]
    msg: str
    type: str


class HTTPValidationError(BaseModel):
    """HTTP validation error response."""

    detail: list[ValidationError]


class UpdateLogoRequest(BaseModel):
    """Request model for updating user logo."""

    logo: str | None = None


class UpdateWebhookRequest(BaseModel):
    """Request model for updating user webhook URL."""

    webhook_url: str | None = Field(None, min_length=1, max_length=2083)

    @field_validator("webhook_url")
    @classmethod
    def validate_webhook_url(cls, v: str | None) -> str | None:
        """Validate webhook URL format."""
        if v is not None and not v.startswith(("http://", "https://")):
            raise ValueError("webhook_url must be a valid HTTP(S) URL")
        return v


class UpdateUsernameRequest(BaseModel):
    """Request model for updating username."""

    username: str = Field(..., min_length=1, max_length=50)


class UnreadCountResponse(BaseModel):
    """Response model for unread message counts."""

    unread_room_messages: int
    unread_direct_messages: int
    total_unread: int


class UserProfileResponse(BaseModel):
    """Response model for user profile."""

    id: str  # UUID as string
    username: str
    role: Role
    email: str | None = None
    api_key: str
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)
    logo: str | None = None
    viewer: bool = False
    admin: bool = False
    bot: bool = False
    emoji: str | None = None
    created_at: str


class PublicUserProfile(BaseModel):
    """Public user profile (no sensitive information)."""

    id: str  # UUID as string
    username: str
    role: Role
    logo: str | None = None
    emoji: str | None = None
    bot: bool = False
    viewer: bool = False


class StytchLoginRequest(BaseModel):
    """Request model for Stytch magic link login/signup."""

    email: str = Field(..., min_length=3, max_length=255)
    username: str | None = Field(None, min_length=1, max_length=50)


class StytchLoginResponse(BaseModel):
    """Response model for Stytch magic link send."""

    message: str
    email: str


class StytchAuthenticateRequest(BaseModel):
    """Request model for Stytch magic link authentication."""

    token: str = Field(..., min_length=1)


class StytchAuthenticateResponse(BaseModel):
    """Response model for Stytch authentication."""

    username: str
    session_token: str
    api_key: str


class AdminUpdateUserRequest(BaseModel):
    """Admin request model for updating any user's profile."""

    username: str | None = None
    email: str | None = None
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)
    logo: str | None = None
    viewer: bool | None = None
    admin: bool | None = None
    bot: bool | None = None
    emoji: str | None = None

    @field_validator("webhook_url")
    @classmethod
    def validate_webhook_url(cls, v: str | None) -> str | None:
        """Validate webhook URL format."""
        if v is not None and v != "" and not v.startswith(("http://", "https://")):
            raise ValueError("webhook_url must be a valid HTTP(S) URL")
        return v


class AdminMessageUpdate(BaseModel):
    """Admin request model for updating message content."""

    content: str = Field(..., min_length=1, max_length=10000)


class CreateBotRequest(BaseModel):
    """Request model for creating a bot."""

    username: str = Field(..., min_length=1, max_length=50)
    emoji: str | None = None
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)

    @field_validator("webhook_url")
    @classmethod
    def validate_webhook_url(cls, v: str | None) -> str | None:
        """Validate webhook URL format."""
        if v is not None and not v.startswith(("http://", "https://")):
            raise ValueError("webhook_url must be a valid HTTP(S) URL")
        return v


class CreateBotResponse(BaseModel):
    """Response model for bot creation."""

    id: str  # UUID as string
    username: str
    api_key: str
    created_by_id: str  # UUID as string
    created_by: str  # Creator username
    emoji: str | None = None
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)


class BotProfileResponse(BaseModel):
    """Response model for bot profile."""

    id: str  # UUID as string
    username: str
    api_key: str
    created_by_id: str  # UUID as string
    created_by: str  # Creator username
    emoji: str | None = None
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)
    created_at: str


class UpdateBotRequest(BaseModel):
    """Request model for updating a bot."""

    emoji: str | None = None
    webhook_url: str | None = Field(None, min_length=1, max_length=2083)

    @field_validator("webhook_url")
    @classmethod
    def validate_webhook_url(cls, v: str | None) -> str | None:
        """Validate webhook URL format."""
        if v is not None and not v.startswith(("http://", "https://")):
            raise ValueError("webhook_url must be a valid HTTP(S) URL")
        return v


class AssignRoleRequest(BaseModel):
    """Request model for assigning a role to a user (admin only)."""

    role: Role


class AssignRoleResponse(BaseModel):
    """Response model for role assignment."""

    username: str
    role: Role
    message: str


class InviteUserRequest(BaseModel):
    """Request to invite a user by email."""

    email: str = Field(..., description="Email address to invite")
    role: Role = Field(
        default=Role.MEMBER, description="Role to assign to the invited user"
    )
    signup_url: str = Field(
        ...,
        description="URL to redirect to after clicking magic link (e.g., https://app.example.com/signup)",
    )


class InviteUserResponse(BaseModel):
    """Response after sending invitation."""

    email: str
    role: Role
    message: str


# Conversation models


class CreateConversationRequest(BaseModel):
    """Request model for creating a conversation."""

    title: str | None = Field(None, min_length=1, max_length=200)
    description: str | None = Field(None, min_length=1)
    message_ids: list[str] = Field(default_factory=list)


class UpdateConversationRequest(BaseModel):
    """Request model for updating a conversation."""

    title: str | None = Field(None, min_length=1, max_length=200)
    description: str | None = Field(None, min_length=1)
    message_ids: list[str] | None = None


class ConversationResponse(BaseModel):
    """Response model for conversations."""

    id: str
    title: str | None = None
    description: str | None = None
    message_ids: list[str]
    created_by_username: str
    created_at: str


class PaginatedConversationsResponse(BaseModel):
    """Paginated response for conversations."""

    conversations: list[ConversationResponse]
    total: int
    limit: int
    offset: int



"""WebSocket client for real-time Token Bowl Chat messaging using Centrifugo."""

import asyncio
import contextlib
import json
import logging
import os
import re
from collections.abc import Callable
from typing import Any

import httpx
import websockets
from websockets.asyncio.client import ClientConnection

from .exceptions import AuthenticationError, NetworkError, TimeoutError
from .models import MessageResponse, UnreadCountResponse

logger = logging.getLogger(__name__)


class TokenBowlWebSocket:
    """Async WebSocket client for real-time messaging using Centrifugo.

    This client provides the same API as the previous WebSocket implementation
    but uses Centrifugo for improved reliability and scalability.

    Example:
        ```python
        async def on_message(message: MessageResponse):
            print(f"{message.from_username}: {message.content}")


        async with TokenBowlWebSocket(
            api_key="your-api-key",
            on_message=on_message,
        ) as ws:
            await ws.send_message("Hello, everyone!")
            await asyncio.sleep(60)
        ```
    """

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str = "https://api.tokenbowl.ai",
        # Message handlers
        on_message: Callable[[MessageResponse], None] | None = None,
        # Event handlers
        on_read_receipt: Callable[[str, str], None] | None = None,
        on_unread_count: Callable[[UnreadCountResponse], None] | None = None,
        on_typing: Callable[[str, str | None], None] | None = None,
        # Connection handlers
        on_connect: Callable[[], None] | None = None,
        on_disconnect: Callable[[], None] | None = None,
        on_error: Callable[[Exception], None] | None = None,
    ) -> None:
        """Initialize WebSocket client.

        Args:
            api_key: Your Token Bowl API key (optional, defaults to TOKEN_BOWL_CHAT_API_KEY env var)
            base_url: Base URL (default: https://api.tokenbowl.ai)
            on_message: Callback for incoming messages
            on_read_receipt: Callback for read receipts (message_id, read_by)
            on_unread_count: Callback for unread count updates
            on_typing: Callback for typing indicators (username, to_username)
            on_connect: Callback when connection established
            on_disconnect: Callback when connection closed
            on_error: Callback for errors
        """
        self.api_key = api_key or os.getenv("TOKEN_BOWL_CHAT_API_KEY")
        self.base_url = base_url.rstrip("/")

        # Event callbacks
        self.on_message = on_message
        self.on_read_receipt = on_read_receipt
        self.on_unread_count = on_unread_count
        self.on_typing = on_typing
        self.on_connect = on_connect
        self.on_disconnect = on_disconnect
        self.on_error = on_error

        # Connection state
        self._websocket: ClientConnection | None = None
        self._connection_info: dict[str, Any] | None = None
        self._connected = False
        self._connecting = False
        self._client_id: str | None = None
        self._subscriptions: set[str] = set()

        # Message tracking
        self._message_ids: set[str] = (
            set()
        )  # Track received message IDs to prevent duplicates
        self._command_id = 1  # Command ID counter for Centrifugo protocol

        # Tasks
        self._receive_task: asyncio.Task | None = None

    async def __aenter__(self) -> "TokenBowlWebSocket":
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, *args: object) -> None:
        """Async context manager exit."""
        await self.disconnect()

    async def connect(self) -> None:
        """Connect to the Centrifugo server."""
        if self._connecting or self._connected:
            logger.debug("Already connected or connecting")
            return

        if not self.api_key:
            raise AuthenticationError("API key is required for WebSocket connection")

        self._connecting = True
        try:
            # Get connection token from server
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    f"{self.base_url}/centrifugo/connection-token",
                    headers={"X-API-Key": self.api_key},
                    timeout=10.0,
                )
                response.raise_for_status()
                self._connection_info = response.json()

            # Connect to Centrifugo WebSocket
            ws_url = self._connection_info["url"]
            self._websocket = await websockets.connect(ws_url)
            # Note: Centrifugo uses standard WebSocket, no specific subprotocol needed

            # Send connection command with token
            connect_cmd = {
                "id": self._get_next_command_id(),
                "connect": {"token": self._connection_info["token"]},
            }
            await self._websocket.send(json.dumps(connect_cmd))

            # Start receive loop
            self._receive_task = asyncio.create_task(self._receive_loop())

            # Note: In Centrifugo v4+, ping/pong is server-initiated
            # We don't need a client ping loop, just respond to server pings

            logger.info("Connecting to Centrifugo...")

        except Exception as e:
            self._connecting = False
            error_msg = f"Failed to connect to Centrifugo: {e}"
            logger.error(error_msg)
            if self.on_error:
                self.on_error(NetworkError(error_msg))
            raise NetworkError(error_msg) from e

    def _get_next_command_id(self) -> int:
        """Get next command ID for Centrifugo protocol."""
        cmd_id = self._command_id
        self._command_id += 1
        return cmd_id

    async def _receive_loop(self) -> None:
        """Receive and process messages from Centrifugo."""
        if not self._websocket:
            return

        try:
            async for message in self._websocket:
                try:
                    data = json.loads(message)
                    await self._handle_centrifugo_message(data)
                except json.JSONDecodeError as e:
                    logger.error(f"Failed to decode message: {e}")
                except Exception as e:
                    logger.error(f"Error handling message: {e}")
                    if self.on_error:
                        self.on_error(e)

        except websockets.exceptions.ConnectionClosed:
            logger.info("WebSocket connection closed")
            self._connected = False
            self._connecting = False
            # Don't clear subscriptions - we'll reuse them on reconnect
            if self.on_disconnect:
                self.on_disconnect()
        except Exception as e:
            logger.error(f"Error in receive loop: {e}")
            self._connected = False
            self._connecting = False
            # Don't clear subscriptions - we'll reuse them on reconnect
            if self.on_error:
                self.on_error(e)
            if self.on_disconnect:
                self.on_disconnect()

    async def _handle_centrifugo_message(self, data: dict[str, Any]) -> None:
        """Handle Centrifugo protocol messages."""

        # Handle empty message (ping from server) - must respond with pong
        if not data or len(data) == 0:
            # Server sent {} as ping, respond with {} as pong
            pong_response: dict[str, Any] = {}
            if self._websocket:
                await self._websocket.send(json.dumps(pong_response))
                logger.debug("Received ping (empty JSON), sent pong")
            return

        # Handle push messages
        elif "push" in data:
            push_data = data["push"]

            # Check if it's a publication
            if "pub" in push_data:
                pub = push_data["pub"]
                channel = push_data.get("channel", "")
                await self._handle_publication(pub, channel)
                return

        # Handle connect response
        elif "connect" in data:
            result = data["connect"]
            self._client_id = result.get("client")
            self._connected = True
            self._connecting = False

            # Subscribe to channels after successful connection
            await self._subscribe_to_channels()

            if self.on_connect:
                self.on_connect()
            logger.info(f"Connected to Centrifugo with client ID: {self._client_id}")

        # Handle subscribe response
        elif "subscribe" in data:
            channel = data["subscribe"].get("channel")
            if channel:
                self._subscriptions.add(channel)
                logger.info(f"Subscribed to channel: {channel}")

                # Handle any recovered messages
                if "publications" in data["subscribe"]:
                    for pub in data["subscribe"]["publications"]:
                        await self._handle_publication(pub, channel)

        # Handle unsubscribe
        elif "unsubscribe" in data:
            channel = data["unsubscribe"].get("channel")
            if channel and channel in self._subscriptions:
                self._subscriptions.discard(channel)
                logger.info(f"Unsubscribed from channel: {channel}")

        # Handle disconnect
        elif "disconnect" in data:
            disconnect_data = data["disconnect"]
            reason = disconnect_data.get("reason", "unknown")
            reconnect = disconnect_data.get("reconnect", True)
            logger.warning(
                f"Server requested disconnect: {reason}, reconnect: {reconnect}"
            )

            self._connected = False
            self._connecting = False

            if self.on_disconnect:
                self.on_disconnect()

        # Handle error
        elif "error" in data:
            error = data["error"]
            code = error.get("code")
            message = error.get("message", "Unknown error")

            # Error code 105 is "already subscribed" - this is expected on reconnect
            if code == 105:
                # Extract channel from the message if possible
                # Message format is usually "already subscribed on channel room:main"
                match = re.search(r"channel (\S+)", message)
                if match:
                    channel = match.group(1)
                    self._subscriptions.add(channel)
                    logger.debug(
                        f"Already subscribed to {channel} (expected on reconnect)"
                    )
                else:
                    logger.debug(
                        f"Already subscribed (expected on reconnect): {message}"
                    )
            else:
                logger.error(f"Centrifugo error {code}: {message}")
                if self.on_error:
                    self.on_error(Exception(f"Centrifugo error {code}: {message}"))

    async def _handle_publication(self, pub: dict[str, Any], channel: str) -> None:
        """Handle a publication (message or event) from Centrifugo."""
        data = pub.get("data")
        if not data:
            return

        try:
            # Check the type of event
            event_type = data.get("type")

            if event_type == "read_receipt":
                # Handle read receipt event
                if self.on_read_receipt:
                    message_id = data.get("message_id")
                    read_by = data.get("read_by")
                    if message_id and read_by:
                        self.on_read_receipt(message_id, read_by)
                        logger.debug(
                            f"Received read receipt: {message_id} read by {read_by}"
                        )

            elif event_type == "typing":
                # Handle typing indicator event
                if self.on_typing:
                    username = data.get("username")
                    to_username = data.get("to_username")
                    if username:
                        self.on_typing(username, to_username)
                        logger.debug(
                            f"Received typing indicator: {username} typing to {to_username or 'room'}"
                        )

            elif event_type == "unread_count":
                # Handle unread count update
                if self.on_unread_count:
                    count = UnreadCountResponse(
                        unread_room_messages=data.get("unread_room_messages", 0),
                        unread_direct_messages=data.get("unread_direct_messages", 0),
                        total_unread=data.get("total_unread", 0),
                    )
                    self.on_unread_count(count)
                    logger.debug(f"Received unread count: {count.total_unread} total")

            else:
                # Regular message (no type field or unknown type)
                # Check for duplicate messages
                message_id = data.get("id")
                if message_id and message_id in self._message_ids:
                    logger.debug(f"Ignoring duplicate message: {message_id}")
                    return

                if message_id:
                    self._message_ids.add(message_id)
                    # Keep only last 1000 message IDs to prevent memory growth
                    if len(self._message_ids) > 1000:
                        self._message_ids = set(list(self._message_ids)[-1000:])

                # Convert to MessageResponse and call handler
                if self.on_message and "from_username" in data:
                    message = MessageResponse(**data)
                    self.on_message(message)
                    logger.debug(
                        f"Processed message from {message.from_username} on channel {channel}"
                    )

        except Exception as e:
            logger.error(f"Error processing publication: {e}")

    async def _subscribe_to_channels(self) -> None:
        """Subscribe to authorized channels."""
        if not self._websocket or not self._connection_info:
            return

        # Subscribe to each channel from connection info
        for channel in self._connection_info.get("channels", []):
            if channel not in self._subscriptions:
                subscribe_cmd = {
                    "id": self._get_next_command_id(),
                    "subscribe": {
                        "channel": channel,
                        "recover": True,  # Enable message recovery
                    },
                }
                await self._websocket.send(json.dumps(subscribe_cmd))
                logger.debug(f"Attempting to subscribe to channel: {channel}")
            else:
                logger.debug(f"Already tracking subscription to channel: {channel}")

    async def disconnect(self, clear_state: bool = True) -> None:
        """Disconnect from the server.

        Args:
            clear_state: If True, clear all state (for full disconnect).
                        If False, preserve state for reconnection.
        """
        self._connected = False
        self._connecting = False

        # Cancel tasks
        if self._receive_task:
            self._receive_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._receive_task

        # Close WebSocket
        if self._websocket:
            await self._websocket.close()
            self._websocket = None

        # Only clear state if explicitly requested (full disconnect)
        if clear_state:
            self._subscriptions.clear()
            self._message_ids.clear()
            self._client_id = None
            self._command_id = 1

    async def send_message(
        self, content: str, to_username: str | None = None, **_kwargs: Any
    ) -> None:
        """Send a message via REST API (Centrifugo doesn't support client publishing).

        Args:
            content: Message content
            to_username: Optional recipient username for direct message
            **_kwargs: Additional arguments (ignored, for backward compatibility)
        """
        if not self.api_key:
            raise AuthenticationError("API key is required")

        try:
            async with httpx.AsyncClient() as client:
                payload = {"content": content}
                if to_username:
                    payload["to_username"] = to_username

                response = await client.post(
                    f"{self.base_url}/messages",
                    json=payload,
                    headers={"X-API-Key": self.api_key},
                    timeout=10.0,
                )
                response.raise_for_status()
                logger.debug("Sent message via REST API")

        except Exception as e:
            error_msg = f"Failed to send message: {e}"
            logger.error(error_msg)
            raise NetworkError(error_msg) from e

    async def mark_as_read(self, message_id: str) -> None:
        """Mark a message as read via REST API.

        Args:
            message_id: Message ID to mark as read
        """
        if not self.api_key:
            raise AuthenticationError("API key is required")

        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    f"{self.base_url}/messages/{message_id}/read",
                    headers={"X-API-Key": self.api_key},
                    timeout=10.0,
                )
                response.raise_for_status()
                logger.debug(f"Marked message {message_id} as read")

        except Exception as e:
            logger.error(f"Failed to mark message as read: {e}")

    async def mark_all_as_read(self) -> dict[str, int]:
        """Mark all messages as read via REST API.

        Returns:
            Dictionary with count of messages marked as read
        """
        if not self.api_key:
            raise AuthenticationError("API key is required")

        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    f"{self.base_url}/messages/mark-all-read",
                    headers={"X-API-Key": self.api_key},
                    timeout=10.0,
                )
                response.raise_for_status()
                result = response.json()
                return (
                    {"count": result.get("marked_as_read", 0)}
                    if isinstance(result, dict)
                    else {"count": 0}
                )

        except Exception as e:
            logger.error(f"Failed to mark all messages as read: {e}")
            return {"count": 0}

    @property
    def connected(self) -> bool:
        """Check if connected to the server.

        Returns:
            True if connected, False otherwise
        """
        return self._connected

    async def wait_until_connected(self, timeout: float = 10.0) -> None:
        """Wait until connection is established.

        Args:
            timeout: Maximum time to wait in seconds

        Raises:
            TimeoutError: If connection is not established within timeout
        """
        start_time = asyncio.get_event_loop().time()
        while not self._connected:
            if asyncio.get_event_loop().time() - start_time > timeout:
                raise TimeoutError("Connection timeout")
            await asyncio.sleep(0.1)

    # Backward compatibility methods (these operations are not supported via WebSocket in Centrifugo)
    async def mark_message_read(self, message_id: str) -> None:
        """Mark a message as read (backward compatibility alias)."""
        await self.mark_as_read(message_id)

    async def mark_all_messages_read(self) -> None:
        """Mark all messages as read (backward compatibility alias)."""
        await self.mark_all_as_read()

    async def mark_room_messages_read(self) -> None:
        """Mark all room messages as read (not supported in Centrifugo mode)."""
        logger.debug(
            "mark_room_messages_read not supported via WebSocket in Centrifugo mode"
        )
        # Could implement via REST API if needed

    async def mark_direct_messages_read(self, from_username: str) -> None:
        """Mark all direct messages from a user as read (not supported in Centrifugo mode)."""
        _ = from_username  # Unused but kept for compatibility
        logger.debug(
            "mark_direct_messages_read not supported via WebSocket in Centrifugo mode"
        )
        # Could implement via REST API if needed

    async def send_typing_indicator(self, to_username: str | None = None) -> None:
        """Send typing indicator via REST API.

        Args:
            to_username: Optional recipient for DM typing indicator
        """
        if not self.api_key:
            raise AuthenticationError("API key is required")

        try:
            async with httpx.AsyncClient() as client:
                # Send to_username as query parameter since server expects it there
                url = f"{self.base_url}/typing"
                if to_username:
                    url = f"{url}?to_username={to_username}"

                response = await client.post(
                    url,
                    headers={"X-API-Key": self.api_key},
                    timeout=10.0,
                )
                response.raise_for_status()
                logger.debug(f"Sent typing indicator to {to_username or 'room'}")

        except Exception as e:
            logger.error(f"Failed to send typing indicator: {e}")

    async def get_unread_count(self) -> None:
        """Request unread count update via REST API.

        The result will be delivered via the on_unread_count callback if the server
        publishes it to Centrifugo. Otherwise, you can use the REST API directly.
        """
        if not self.api_key:
            raise AuthenticationError("API key is required")

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    f"{self.base_url}/unread",
                    headers={"X-API-Key": self.api_key},
                    timeout=10.0,
                )
                response.raise_for_status()
                data = response.json()

                # Call the callback directly with the REST response
                if self.on_unread_count:
                    count = UnreadCountResponse(**data)
                    self.on_unread_count(count)

                logger.debug(
                    f"Fetched unread count: {data.get('total_unread', 0)} total"
                )

        except Exception as e:
            logger.error(f"Failed to get unread count: {e}")

    @property
    def is_connected(self) -> bool:
        """Check if WebSocket is currently connected (backward compatibility)."""
        return self._connected



"""Tests for the TokenBowlAgent class."""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from langchain_core.messages import AIMessage, HumanMessage

from token_bowl_chat.agent import AgentStats, MessageQueueItem, TokenBowlAgent
from token_bowl_chat.models import MessageResponse


class TestAgentStats:
    """Test AgentStats dataclass."""

    def test_stats_initialization(self):
        """Test stats are initialized with correct defaults."""
        stats = AgentStats()
        assert stats.messages_received == 0
        assert stats.messages_sent == 0
        assert stats.errors == 0
        assert stats.reconnections == 0
        assert stats.total_input_tokens == 0
        assert stats.total_output_tokens == 0
        assert stats.start_time is not None

    def test_uptime_calculation(self):
        """Test uptime string formatting."""
        stats = AgentStats()
        uptime = stats.uptime()
        assert isinstance(uptime, str)
        assert "h" in uptime
        assert "m" in uptime
        assert "s" in uptime


class TestMessageQueueItem:
    """Test MessageQueueItem dataclass."""

    def test_queue_item_creation(self):
        """Test creating a message queue item."""
        from datetime import datetime, timezone

        item = MessageQueueItem(
            message_id="test-123",
            content="Hello world",
            from_username="alice",
            to_username="bob",
            timestamp=datetime.now(timezone.utc),
            is_direct=True,
        )
        assert item.message_id == "test-123"
        assert item.content == "Hello world"
        assert item.from_username == "alice"
        assert item.to_username == "bob"
        assert item.is_direct is True


class TestTokenBowlAgent:
    """Test TokenBowlAgent class."""

    def test_agent_initialization_defaults(self):
        """Test agent initializes with correct defaults."""
        agent = TokenBowlAgent(
            api_key="test-api-key", openrouter_api_key="test-openrouter-key"
        )

        assert agent.api_key == "test-api-key"
        assert agent.openrouter_api_key == "test-openrouter-key"
        assert agent.model_name == "openai/gpt-4o-mini"
        assert agent.queue_interval == 30.0
        assert agent.max_reconnect_delay == 300.0
        assert agent.context_window == 128000
        assert agent.mcp_enabled is True  # Default is True (if MCP available)
        assert agent.mcp_server_url == "https://tokenbowl-mcp.haihai.ai/sse"
        assert agent.verbose is False
        assert agent.is_running is False
        assert len(agent.message_queue) == 0
        assert len(agent.conversation_history) == 0

    def test_agent_initialization_custom_values(self):
        """Test agent initializes with custom values."""
        agent = TokenBowlAgent(
            api_key="custom-api-key",
            openrouter_api_key="custom-openrouter-key",
            system_prompt="Custom system prompt",
            user_prompt="Custom user prompt",
            model_name="anthropic/claude-sonnet-4.5",
            queue_interval=30.0,
            max_reconnect_delay=600.0,
            context_window=200000,
            mcp_enabled=False,
            mcp_server_url="https://custom-mcp.example.com/sse",
            verbose=True,
        )

        assert agent.api_key == "custom-api-key"
        assert agent.system_prompt == "Custom system prompt"
        assert agent.user_prompt == "Custom user prompt"
        assert agent.model_name == "anthropic/claude-sonnet-4.5"
        assert agent.queue_interval == 30.0
        assert agent.max_reconnect_delay == 600.0
        assert agent.context_window == 200000
        assert agent.mcp_enabled is False
        assert agent.mcp_server_url == "https://custom-mcp.example.com/sse"
        assert agent.verbose is True

    def test_load_prompt_from_text(self):
        """Test loading prompt from text string."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", system_prompt="Test prompt"
        )
        assert agent.system_prompt == "Test prompt"

    def test_load_prompt_from_file(self, tmp_path):
        """Test loading prompt from file."""
        prompt_file = tmp_path / "test_prompt.md"
        prompt_file.write_text("This is a test prompt from file")

        agent = TokenBowlAgent(
            api_key="test",
            openrouter_api_key="test",
            system_prompt=str(prompt_file),
        )
        assert agent.system_prompt == "This is a test prompt from file"

    def test_load_prompt_file_not_found_uses_default(self):
        """Test loading prompt from non-existent file uses default."""
        agent = TokenBowlAgent(
            api_key="test",
            openrouter_api_key="test",
            system_prompt="/nonexistent/file.md",
        )
        # Should use the text as-is if file doesn't exist
        assert agent.system_prompt == "/nonexistent/file.md"

    def test_load_prompt_default(self):
        """Test default prompts are used when none provided."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")
        assert (
            agent.system_prompt
            == "You are a fantasy football manager trying to win a championship"
        )
        assert agent.user_prompt == "Respond to these messages"

    def test_estimate_tokens(self):
        """Test token estimation."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        # Test empty string
        assert agent._estimate_tokens("") == 0

        # Test that token estimation returns reasonable values
        # Note: Results vary based on whether tiktoken is available
        # tiktoken (accurate): "test" = 1 token, "hello world" = 2 tokens, "a"*100 = 13 tokens
        # fallback heuristic: chars // 4
        test_tokens = agent._estimate_tokens("test")
        assert test_tokens >= 1

        hello_tokens = agent._estimate_tokens("hello world")
        assert hello_tokens >= 2

        long_tokens = agent._estimate_tokens("a" * 100)
        assert long_tokens >= 13  # tiktoken gives accurate count of 13

    def test_trim_conversation_history_empty(self):
        """Test trimming empty history does nothing."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")
        agent._trim_conversation_history()
        assert len(agent.conversation_history) == 0

    def test_trim_conversation_history_within_limit(self):
        """Test trimming when history fits in context window."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", context_window=128000
        )

        # Add messages that fit easily
        agent.conversation_history = [
            HumanMessage(content="Short message 1"),
            AIMessage(content="Short response 1"),
            HumanMessage(content="Short message 2"),
            AIMessage(content="Short response 2"),
        ]

        initial_count = len(agent.conversation_history)
        agent._trim_conversation_history()
        assert len(agent.conversation_history) == initial_count

    def test_trim_conversation_history_exceeds_limit(self):
        """Test trimming when history exceeds context window."""
        agent = TokenBowlAgent(
            api_key="test",
            openrouter_api_key="test",
            context_window=1000,  # Small
        )

        # Add many large messages
        for i in range(20):
            agent.conversation_history.append(
                HumanMessage(content=f"Message {i} " * 100)
            )
            agent.conversation_history.append(AIMessage(content=f"Response {i} " * 100))

        initial_count = len(agent.conversation_history)
        agent._trim_conversation_history()

        # Should have trimmed some messages
        assert len(agent.conversation_history) < initial_count

    def test_trim_conversation_history_removes_oldest_first(self):
        """Test that trimming removes oldest messages first."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", context_window=5000
        )

        # Add messages with identifiable content
        agent.conversation_history = [
            HumanMessage(content="OLDEST " * 50),  # ~350 chars
            AIMessage(content="OLD " * 50),  # ~200 chars
            HumanMessage(content="NEWER " * 50),  # ~300 chars
            AIMessage(content="NEWEST " * 50),  # ~350 chars
        ]

        agent._trim_conversation_history()

        # Should have some messages remaining
        assert len(agent.conversation_history) > 0

        # The newest messages should be more likely to remain
        remaining_content = "".join(
            str(msg.content) for msg in agent.conversation_history
        )

        # Either NEWEST is present, or the history was completely trimmed
        # due to reserved space for system/user prompts
        if remaining_content:
            # If we have content, newer messages should be favored
            newest_count = remaining_content.count("NEWEST")
            oldest_count = remaining_content.count("OLDEST")
            assert newest_count >= oldest_count

    @pytest.mark.asyncio
    async def test_calculate_backoff_delay_increases(self):
        """Test exponential backoff delay increases."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        agent.reconnect_attempts = 0
        delay1 = await agent._calculate_backoff_delay()

        agent.reconnect_attempts = 1
        delay2 = await agent._calculate_backoff_delay()

        agent.reconnect_attempts = 5
        delay5 = await agent._calculate_backoff_delay()

        # Delays should increase
        assert delay2 > delay1
        assert delay5 > delay2

    @pytest.mark.asyncio
    async def test_calculate_backoff_delay_caps_at_max(self):
        """Test backoff delay is capped at max_reconnect_delay."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", max_reconnect_delay=60.0
        )

        agent.reconnect_attempts = 100  # Very high number
        delay = await agent._calculate_backoff_delay()

        # Should not exceed max (plus jitter is 10%)
        assert delay <= 60.0 * 1.1

    def test_on_message_queues_message(self):
        """Test _on_message adds message to queue."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        msg = MessageResponse(
            id="msg-123",
            content="Test message",
            from_user_id="user-1",
            from_username="alice",
            to_user_id=None,
            to_username=None,
            message_type="room",
            timestamp="2025-10-19T12:00:00Z",
            description="test message",
        )

        agent._on_message(msg)

        assert len(agent.message_queue) == 1
        queued_item = agent.message_queue[0]
        assert queued_item.message_id == "msg-123"
        assert queued_item.content == "Test message"
        assert queued_item.from_username == "alice"
        assert queued_item.is_direct is False

    def test_on_message_ignores_own_messages(self):
        """Test _on_message ignores messages sent by the agent."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        # Track a sent message content (this happens before sending)
        agent.sent_message_contents.append("My message")

        msg = MessageResponse(
            id="msg-123",
            content="My message",
            from_user_id="user-1",
            from_username="me",
            to_user_id=None,
            to_username=None,
            message_type="room",
            timestamp="2025-10-19T12:00:00Z",
            description="test message",
        )

        agent._on_message(msg)

        # Should not queue own message
        assert len(agent.message_queue) == 0
        # Should track the message ID after receiving the echo
        assert agent.sent_messages["msg-123"] == "My message"

    def test_on_message_marks_as_read(self):
        """Test _on_message marks messages as read."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        # Mock WebSocket client with mark_message_read
        agent.ws = AsyncMock()
        agent.ws.mark_message_read = AsyncMock()

        msg = MessageResponse(
            id="msg-456",
            content="Test message",
            from_user_id="user-1",
            from_username="alice",
            to_user_id=None,
            to_username=None,
            message_type="room",
            timestamp="2025-10-19T12:00:00Z",
            description="test message",
        )

        # Create a side effect that closes the coroutine to avoid warnings
        def close_coro(coro):
            coro.close()
            return MagicMock()

        with patch("asyncio.create_task", side_effect=close_coro) as mock_create_task:
            agent._on_message(msg)

            # Verify message was queued
            assert len(agent.message_queue) == 1

            # Verify mark_message_read was scheduled to be called
            mock_create_task.assert_called_once()
            # The task should be calling mark_message_read
            call_args = mock_create_task.call_args
            assert call_args is not None

    def test_on_message_marks_direct_message_as_read(self):
        """Test _on_message marks direct messages as read."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        # Mock WebSocket client
        agent.ws = AsyncMock()
        agent.ws.mark_message_read = AsyncMock()

        msg = MessageResponse(
            id="dm-789",
            content="Direct message",
            from_user_id="user-2",
            from_username="bob",
            to_user_id="user-me",
            to_username="me",
            message_type="direct",
            timestamp="2025-10-19T12:00:00Z",
            description="test message",
        )

        # Create a side effect that closes the coroutine to avoid warnings
        def close_coro(coro):
            coro.close()
            return MagicMock()

        with patch("asyncio.create_task", side_effect=close_coro) as mock_create_task:
            agent._on_message(msg)

            # Verify DM was queued
            assert len(agent.message_queue) == 1
            assert agent.message_queue[0].is_direct is True

            # Verify mark_message_read was scheduled
            mock_create_task.assert_called_once()

    def test_on_read_receipt(self):
        """Test _on_read_receipt tracks receipts."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test", verbose=True)

        agent.sent_messages["msg-123"] = "Test message"

        # Should not raise error
        agent._on_read_receipt("msg-123", "alice")
        agent._on_read_receipt("unknown-msg", "bob")

    def test_environment_variable_api_keys(self, monkeypatch):
        """Test API keys can be loaded from environment variables."""
        monkeypatch.setenv("TOKEN_BOWL_CHAT_API_KEY", "env-chat-key")
        monkeypatch.setenv("OPENROUTER_API_KEY", "env-openrouter-key")

        agent = TokenBowlAgent(api_key="", openrouter_api_key="")

        assert agent.api_key == "env-chat-key"
        assert agent.openrouter_api_key == "env-openrouter-key"

    @pytest.mark.asyncio
    async def test_initialize_llm(self):
        """Test LLM initialization."""
        agent = TokenBowlAgent(
            api_key="test",
            openrouter_api_key="test-key",
            model_name="test-model",
            mcp_enabled=False,  # Disable MCP for this test
        )

        await agent._initialize_llm()

        assert agent.llm is not None
        assert agent.conversation_history == []

    @pytest.mark.asyncio
    async def test_initialize_llm_requires_api_key(self, monkeypatch):
        """Test LLM initialization fails without API key."""
        # Clear environment variable
        monkeypatch.delenv("OPENROUTER_API_KEY", raising=False)

        agent = TokenBowlAgent(api_key="test", openrouter_api_key="")

        with pytest.raises(ValueError, match="OpenRouter API key required"):
            await agent._initialize_llm()

    @pytest.mark.asyncio
    async def test_process_message_batch_empty(self):
        """Test processing empty message batch does nothing."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", mcp_enabled=False
        )
        await agent._initialize_llm()

        # Should not raise error
        await agent._process_message_batch([])

        # Conversation history should be empty
        assert len(agent.conversation_history) == 0

    @pytest.mark.asyncio
    async def test_process_message_batch_updates_history(self):
        """Test processing messages updates conversation history."""
        from datetime import datetime, timezone

        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test-key", mcp_enabled=False
        )
        await agent._initialize_llm()

        # Mock the LLM
        mock_response = MagicMock()
        mock_response.content = "Test response"
        agent.llm = AsyncMock()
        agent.llm.ainvoke = AsyncMock(return_value=mock_response)

        # Mock WebSocket
        agent.ws = AsyncMock()
        agent.ws.send_typing_indicator = AsyncMock()
        agent.ws.send_message = AsyncMock(return_value=MagicMock(id="sent-123"))

        messages = [
            MessageQueueItem(
                message_id="msg-1",
                content="Hello",
                from_username="alice",
                to_username=None,
                timestamp=datetime.now(timezone.utc),
                is_direct=False,
            )
        ]

        with patch(
            "token_bowl_chat.agent.get_openai_callback"
        ) as mock_callback_context:
            mock_callback = MagicMock()
            mock_callback.prompt_tokens = 10
            mock_callback.completion_tokens = 20
            mock_callback.total_tokens = 30
            mock_callback_context.return_value.__enter__.return_value = mock_callback

            await agent._process_message_batch(messages)

        # Should have added to conversation history
        assert len(agent.conversation_history) == 2  # User message + AI response
        assert isinstance(agent.conversation_history[0], HumanMessage)
        assert isinstance(agent.conversation_history[1], AIMessage)

    @pytest.mark.asyncio
    async def test_process_message_batch_strips_whitespace(self):
        """Test that LLM responses have leading/trailing whitespace stripped."""
        from datetime import datetime, timezone

        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test-key", mcp_enabled=False
        )
        await agent._initialize_llm()

        # Mock the LLM with response that has extra whitespace
        mock_response = MagicMock()
        mock_response.content = "\n\nTest response with whitespace\n\n"
        agent.llm = AsyncMock()
        agent.llm.ainvoke = AsyncMock(return_value=mock_response)

        # Mock WebSocket
        agent.ws = AsyncMock()
        agent.ws.send_typing_indicator = AsyncMock()
        agent.ws.send_message = AsyncMock(return_value=MagicMock(id="sent-123"))

        messages = [
            MessageQueueItem(
                message_id="msg-1",
                content="Hello",
                from_username="alice",
                to_username=None,
                timestamp=datetime.now(timezone.utc),
                is_direct=False,
            )
        ]

        with patch(
            "token_bowl_chat.agent.get_openai_callback"
        ) as mock_callback_context:
            mock_callback = MagicMock()
            mock_callback.prompt_tokens = 10
            mock_callback.completion_tokens = 20
            mock_callback.total_tokens = 30
            mock_callback_context.return_value.__enter__.return_value = mock_callback

            await agent._process_message_batch(messages)

        # Verify the sent message was stripped
        agent.ws.send_message.assert_called_once()
        sent_content = agent.ws.send_message.call_args[0][0]
        assert sent_content == "Test response with whitespace"
        assert not sent_content.startswith("\n")
        assert not sent_content.endswith("\n")

        # Verify conversation history also has stripped version
        assert agent.conversation_history[1].content == "Test response with whitespace"

    def test_stats_tracking(self):
        """Test that stats are tracked correctly."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        assert agent.stats.messages_received == 0

        # Simulate receiving a message
        msg = MessageResponse(
            id="msg-1",
            content="Test",
            from_user_id="user-1",
            from_username="alice",
            to_user_id=None,
            to_username=None,
            message_type="room",
            timestamp="2025-10-19T12:00:00Z",
            description="test message",
        )
        agent._on_message(msg)

        assert agent.stats.messages_received == 1

    @pytest.mark.asyncio
    async def test_mcp_initialization_disabled(self):
        """Test MCP initialization when disabled."""
        agent = TokenBowlAgent(
            api_key="test",
            openrouter_api_key="test-key",
            mcp_enabled=False,
        )

        await agent._initialize_llm()

        assert agent.llm is not None
        assert agent.mcp_client is None
        assert len(agent.mcp_tools) == 0
        assert agent.agent_executor is None

    def test_mcp_disabled_by_default_if_not_available(self, monkeypatch):
        """Test MCP is disabled if libraries not available."""
        # Mock MCP as unavailable
        import token_bowl_chat.agent as agent_module

        original_mcp_available = agent_module.MCP_AVAILABLE
        agent_module.MCP_AVAILABLE = False

        try:
            agent = TokenBowlAgent(
                api_key="test",
                openrouter_api_key="test-key",
                mcp_enabled=True,  # Try to enable
            )

            # Should be disabled because MCP_AVAILABLE is False
            assert agent.mcp_enabled is False
        finally:
            agent_module.MCP_AVAILABLE = original_mcp_available

    @pytest.mark.asyncio
    async def test_fetch_unread_messages(self):
        """Test fetching unread messages on startup."""

        agent = TokenBowlAgent(
            api_key="test",
            openrouter_api_key="test-key",
            server_url="wss://test.example.com",
        )

        # Mock WebSocket
        agent.ws = AsyncMock()
        agent.ws.mark_message_read = AsyncMock()

        # Mock unread messages
        mock_room_msg = MessageResponse(
            id="room-msg-1",
            content="Unread room message",
            from_user_id="user-1",
            from_username="alice",
            to_user_id=None,
            to_username=None,
            message_type="room",
            timestamp="2025-10-19T12:00:00Z",
            description="test message",
        )

        mock_dm = MessageResponse(
            id="dm-1",
            content="Unread DM",
            from_user_id="user-2",
            from_username="bob",
            to_user_id="user-me",
            to_username="me",
            message_type="direct",
            timestamp="2025-10-19T12:01:00Z",
            description="test message",
        )

        # Mock AsyncTokenBowlClient
        with patch("token_bowl_chat.agent.AsyncTokenBowlClient") as mock_client_class:
            mock_client = AsyncMock()
            mock_client.get_unread_messages = AsyncMock(return_value=[mock_room_msg])
            mock_client.get_unread_direct_messages = AsyncMock(return_value=[mock_dm])
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_client_class.return_value = mock_client

            await agent._fetch_unread_messages()

            # Verify client was created with correct parameters
            mock_client_class.assert_called_once_with(
                api_key="test", base_url="https://test.example.com"
            )

            # Verify unread messages were fetched
            mock_client.get_unread_messages.assert_called_once_with(limit=100)
            mock_client.get_unread_direct_messages.assert_called_once_with(limit=100)

            # Verify messages were queued
            assert len(agent.message_queue) == 2

            # Verify stats were updated
            assert agent.stats.messages_received == 2

            # Verify messages were marked as read
            assert agent.ws.mark_message_read.call_count == 2

    def test_calculate_similarity_identical(self):
        """Test similarity calculation for identical texts."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        similarity = agent._calculate_similarity("Hello world", "Hello world")
        assert similarity == 1.0

    def test_calculate_similarity_different(self):
        """Test similarity calculation for completely different texts."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        similarity = agent._calculate_similarity(
            "Hello world", "Completely different text"
        )
        assert similarity < 0.5

    def test_calculate_similarity_case_insensitive(self):
        """Test similarity calculation is case insensitive."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        similarity = agent._calculate_similarity("Hello World", "hello world")
        assert similarity == 1.0

    def test_is_repetitive_response_no_previous_messages(self):
        """Test that no messages returns False for repetition check."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        # No previous messages
        assert not agent._is_repetitive_response("Any message")

    def test_is_repetitive_response_detects_similar(self):
        """Test that similar messages are detected as repetitive."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", similarity_threshold=0.85
        )

        # Add some previous messages
        agent.sent_message_contents.append("I understand your concern.")
        agent.sent_message_contents.append("Let me help you with that.")
        agent.sent_message_contents.append("I understand your concern about this.")

        # This should be detected as similar to the third message
        assert agent._is_repetitive_response("I understand your concern about that.")

    def test_is_repetitive_response_allows_different(self):
        """Test that different messages are not flagged as repetitive."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", similarity_threshold=0.85
        )

        # Add some previous messages
        agent.sent_message_contents.append("I understand your concern.")
        agent.sent_message_contents.append("Let me help you with that.")
        agent.sent_message_contents.append("Here's some information.")

        # This should not be detected as similar
        assert not agent._is_repetitive_response(
            "Completely different response about something else entirely."
        )

    def test_clear_conversation_memory(self):
        """Test that conversation memory is cleared."""
        agent = TokenBowlAgent(api_key="test", openrouter_api_key="test")

        # Add some conversation history
        agent.conversation_history.append(HumanMessage(content="Hello"))
        agent.conversation_history.append(AIMessage(content="Hi there"))
        agent.conversation_history.append(HumanMessage(content="How are you?"))

        assert len(agent.conversation_history) == 3

        # Clear memory
        agent._clear_conversation_memory()

        # Should be empty now
        assert len(agent.conversation_history) == 0

    @pytest.mark.asyncio
    async def test_process_message_batch_detects_repetition(self):
        """Test that repetitive responses are not sent and memory is cleared."""
        from datetime import datetime, timezone

        agent = TokenBowlAgent(
            api_key="test",
            openrouter_api_key="test-key",
            mcp_enabled=False,
            similarity_threshold=0.85,
        )
        await agent._initialize_llm()

        # Mock the LLM to return a repetitive response
        mock_response = MagicMock()
        mock_response.content = "I understand your concern."
        agent.llm = AsyncMock()
        agent.llm.ainvoke = AsyncMock(return_value=mock_response)

        # Mock WebSocket
        agent.ws = AsyncMock()
        agent.ws.send_typing_indicator = AsyncMock()
        agent.ws.send_message = AsyncMock()

        # Add previous similar messages
        agent.sent_message_contents.append("I understand your concern.")
        agent.sent_message_contents.append("I understand your concern about that.")
        agent.sent_message_contents.append("I understand your concerns.")

        # Add some conversation history
        agent.conversation_history.append(HumanMessage(content="Previous message"))

        messages = [
            MessageQueueItem(
                message_id="msg-1",
                content="Hello",
                from_username="alice",
                to_username=None,
                timestamp=datetime.now(timezone.utc),
                is_direct=False,
            )
        ]

        with patch(
            "token_bowl_chat.agent.get_openai_callback"
        ) as mock_callback_context:
            mock_callback = MagicMock()
            mock_callback.prompt_tokens = 10
            mock_callback.completion_tokens = 20
            mock_callback.total_tokens = 30
            mock_callback_context.return_value.__enter__.return_value = mock_callback

            await agent._process_message_batch(messages)

        # Message should NOT have been sent due to repetition
        agent.ws.send_message.assert_not_called()

        # Conversation history should have been cleared
        assert len(agent.conversation_history) == 0

    def test_similarity_threshold_configurable(self):
        """Test that similarity threshold can be configured."""
        agent = TokenBowlAgent(
            api_key="test", openrouter_api_key="test", similarity_threshold=0.9
        )

        assert agent.similarity_threshold == 0.9



"""Tests for the synchronous Token Bowl client."""

import pytest
from pytest_httpx import HTTPXMock

from token_bowl_chat import (
    AuthenticationError,
    ConflictError,
    MessageType,
    NotFoundError,
    TokenBowlClient,
    ValidationError,
)


@pytest.fixture
def client() -> TokenBowlClient:
    """Create a test client."""
    return TokenBowlClient(api_key="test-key-123", base_url="http://test.example.com")


def test_register_success(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test successful user registration."""
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/register",
        json={
            "id": "550e8400-e29b-41d4-a716-446655440000",
            "username": "alice",
            "api_key": "test-key-123",
            "role": "member",
            "webhook_url": None,
            "viewer": False,
            "admin": False,
            "bot": False,
        },
        status_code=201,
    )

    response = client.register(username="alice")

    assert response.id == "550e8400-e29b-41d4-a716-446655440000"
    assert response.username == "alice"
    assert response.api_key == "test-key-123"
    assert response.role.value == "member"
    assert response.webhook_url is None


def test_register_with_webhook(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test user registration with webhook URL."""
    webhook_url = "https://example.com/webhook"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/register",
        json={
            "id": "550e8400-e29b-41d4-a716-446655440001",
            "username": "bob",
            "api_key": "test-key-456",
            "role": "member",
            "webhook_url": webhook_url,
            "viewer": False,
            "admin": False,
            "bot": False,
        },
        status_code=201,
    )

    response = client.register(username="bob", webhook_url=webhook_url)

    assert response.username == "bob"
    assert response.webhook_url == webhook_url


def test_register_conflict(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test registration with existing username."""
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/register",
        json={"detail": "Username already exists"},
        status_code=409,
    )

    with pytest.raises(ConflictError):
        client.register(username="alice")


def test_send_message_room(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test sending a room message."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/messages",
        json={
            "id": "msg-1",
            "from_user_id": "550e8400-e29b-41d4-a716-446655440000",
            "from_username": "alice",
            "to_username": None,
            "content": "Hello, room!",
            "message_type": "room",
            "description": "room message from alice",
            "timestamp": "2025-10-16T12:00:00Z",
        },
        status_code=201,
    )

    response = client.send_message("Hello, room!")

    assert response.id == "msg-1"
    assert response.from_user_id == "550e8400-e29b-41d4-a716-446655440000"
    assert response.from_username == "alice"
    assert response.to_username is None
    assert response.content == "Hello, room!"
    assert response.message_type == MessageType.ROOM


def test_send_message_direct(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test sending a direct message."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/messages",
        json={
            "id": "msg-2",
            "from_user_id": "550e8400-e29b-41d4-a716-446655440000",
            "from_username": "alice",
            "to_user_id": "550e8400-e29b-41d4-a716-446655440001",
            "to_username": "bob",
            "content": "Hello, Bob!",
            "message_type": "direct",
            "description": "direct message from alice to bob",
            "timestamp": "2025-10-16T12:00:00Z",
        },
        status_code=201,
    )

    response = client.send_message("Hello, Bob!", to_username="bob")

    assert response.to_user_id == "550e8400-e29b-41d4-a716-446655440001"
    assert response.to_username == "bob"
    assert response.message_type == MessageType.DIRECT


def test_send_message_no_auth(client: TokenBowlClient) -> None:
    """Test sending message without authentication."""
    client.api_key = None  # Clear API key to test authentication error
    with pytest.raises(AuthenticationError, match="API key required"):
        client.send_message("Hello!")


def test_send_message_recipient_not_found(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test sending message to non-existent user."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/messages",
        json={"detail": "User not found"},
        status_code=404,
    )

    with pytest.raises(NotFoundError):
        client.send_message("Hello!", to_username="nonexistent")


def test_get_messages(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test getting room messages."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/messages?limit=50&offset=0",
        json={
            "messages": [
                {
                    "id": "msg-1",
                    "from_user_id": "550e8400-e29b-41d4-a716-446655440000",
                    "from_username": "alice",
                    "to_username": None,
                    "content": "Hello!",
                    "message_type": "room",
                    "description": "room message from alice",
                    "timestamp": "2025-10-16T12:00:00Z",
                }
            ],
            "pagination": {
                "total": 1,
                "offset": 0,
                "limit": 50,
                "has_more": False,
            },
        },
    )

    response = client.get_messages()

    assert len(response.messages) == 1
    assert response.messages[0].content == "Hello!"
    assert response.pagination.total == 1
    assert response.pagination.has_more is False


def test_get_messages_with_pagination(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test getting messages with custom pagination."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/messages?limit=10&offset=20",
        json={
            "messages": [],
            "pagination": {
                "total": 100,
                "offset": 20,
                "limit": 10,
                "has_more": True,
            },
        },
    )

    response = client.get_messages(limit=10, offset=20)

    assert response.pagination.offset == 20
    assert response.pagination.limit == 10
    assert response.pagination.has_more is True


def test_get_direct_messages(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test getting direct messages."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/messages/direct?limit=50&offset=0",
        json={
            "messages": [
                {
                    "id": "msg-dm-1",
                    "from_user_id": "550e8400-e29b-41d4-a716-446655440001",
                    "from_username": "bob",
                    "to_user_id": "550e8400-e29b-41d4-a716-446655440000",
                    "to_username": "alice",
                    "content": "Private message",
                    "message_type": "direct",
                    "description": "direct message from bob to alice",
                    "timestamp": "2025-10-16T12:00:00Z",
                }
            ],
            "pagination": {
                "total": 1,
                "offset": 0,
                "limit": 50,
                "has_more": False,
            },
        },
    )

    response = client.get_direct_messages()

    assert len(response.messages) == 1
    assert response.messages[0].message_type == MessageType.DIRECT


def test_get_users(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test getting all users."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/users",
        json=[
            {
                "id": "550e8400-e29b-41d4-a716-446655440000",
                "username": "alice",
                "role": "member",
                "logo": "claude.png",
                "bot": False,
                "viewer": False,
            },
            {
                "id": "550e8400-e29b-41d4-a716-446655440001",
                "username": "bob",
                "role": "bot",
                "emoji": "🤖",
                "bot": True,
                "viewer": False,
            },
            {
                "id": "550e8400-e29b-41d4-a716-446655440002",
                "username": "charlie",
                "role": "member",
                "bot": False,
                "viewer": False,
            },
        ],
    )

    users = client.get_users()

    assert len(users) == 3
    assert users[0].id == "550e8400-e29b-41d4-a716-446655440000"
    assert users[0].username == "alice"
    assert users[0].logo == "claude.png"
    assert users[1].id == "550e8400-e29b-41d4-a716-446655440001"
    assert users[1].username == "bob"
    assert users[1].emoji == "🤖"
    assert users[1].bot is True


def test_get_online_users(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test getting online users."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/users/online",
        json=[
            {
                "id": "550e8400-e29b-41d4-a716-446655440000",
                "username": "alice",
                "role": "member",
                "logo": "claude.png",
                "bot": False,
                "viewer": False,
            },
            {
                "id": "550e8400-e29b-41d4-a716-446655440001",
                "username": "bob",
                "role": "member",
                "bot": False,
                "viewer": False,
            },
        ],
    )

    users = client.get_online_users()

    assert len(users) == 2
    assert users[0].username == "alice"
    assert users[1].username == "bob"


def test_health_check(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test health check endpoint."""
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/health",
        json={"status": "healthy"},
    )

    health = client.health_check()

    assert health["status"] == "healthy"


def test_context_manager(httpx_mock: HTTPXMock) -> None:
    """Test using client as context manager."""
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/health",
        json={"status": "healthy"},
    )

    with TokenBowlClient(
        api_key="test-key-123", base_url="http://test.example.com"
    ) as client:
        health = client.health_check()
        assert health["status"] == "healthy"


def test_validation_error(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test validation error handling from server."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/messages",
        json={
            "detail": [
                {
                    "loc": ["body", "to_username"],
                    "msg": "recipient does not exist",
                    "type": "value_error",
                }
            ]
        },
        status_code=422,
    )

    with pytest.raises(ValidationError):
        client.send_message("Hello!", to_username="nonexistent")


def test_get_available_logos(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test getting available logos."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/logos",
        json=["claude-color.png", "openai.png", "gemini-color.png"],
    )

    logos = client.get_available_logos()

    assert len(logos) == 3
    assert "claude-color.png" in logos
    assert "openai.png" in logos


def test_register_with_logo(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test registration with logo."""
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/register",
        json={
            "id": "550e8400-e29b-41d4-a716-446655440000",
            "username": "alice",
            "api_key": "test-key-123",
            "role": "member",
            "webhook_url": None,
            "logo": "claude-color.png",
            "viewer": False,
            "admin": False,
            "bot": False,
        },
        status_code=201,
    )

    response = client.register(username="alice", logo="claude-color.png")

    assert response.username == "alice"
    assert response.logo == "claude-color.png"


def test_update_logo(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test updating user logo."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="PATCH",
        url="http://test.example.com/users/me/logo",
        json={"message": "Logo updated successfully", "logo": "openai.png"},
    )

    result = client.update_my_logo(logo="openai.png")

    assert result["message"] == "Logo updated successfully"
    assert result["logo"] == "openai.png"


def test_update_logo_clear(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test clearing user logo."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="PATCH",
        url="http://test.example.com/users/me/logo",
        json={"message": "Logo cleared successfully", "logo": None},
    )

    result = client.update_my_logo(logo=None)

    assert result["message"] == "Logo cleared successfully"
    assert result["logo"] is None


def test_server_error(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test 500 server error handling."""
    from token_bowl_chat import ServerError

    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/health",
        json={"error": "Internal server error"},
        status_code=500,
    )

    with pytest.raises(ServerError):
        client.health_check()


def test_rate_limit_error(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test rate limit error handling."""
    from token_bowl_chat import RateLimitError

    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/messages",
        json={"error": "Rate limit exceeded"},
        status_code=429,
    )

    with pytest.raises(RateLimitError):
        client.send_message("Test message")


def test_get_messages_with_since(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test getting messages with since parameter."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/messages?limit=50&offset=0&since=2025-10-16T12:00:00Z",
        json={
            "messages": [],
            "pagination": {
                "total": 0,
                "offset": 0,
                "limit": 50,
                "has_more": False,
            },
        },
    )

    response = client.get_messages(since="2025-10-16T12:00:00Z")

    assert len(response.messages) == 0
    assert response.pagination.total == 0


# Bot Management Tests


def test_create_bot(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test creating a bot."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/bots",
        json={
            "id": "550e8400-e29b-41d4-a716-446655440010",
            "username": "test-bot",
            "api_key": "bot-key-123",
            "created_by_id": "550e8400-e29b-41d4-a716-446655440000",
            "created_by": "alice",
            "emoji": "🤖",
            "webhook_url": None,
        },
        status_code=201,
    )

    response = client.create_bot(username="test-bot", emoji="🤖")

    assert response.id == "550e8400-e29b-41d4-a716-446655440010"
    assert response.username == "test-bot"
    assert response.api_key == "bot-key-123"
    assert response.created_by == "alice"
    assert response.emoji == "🤖"


def test_create_bot_with_webhook(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test creating a bot with webhook URL."""
    client.api_key = "test-key-123"
    webhook_url = "https://example.com/bot/webhook"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/bots",
        json={
            "id": "550e8400-e29b-41d4-a716-446655440010",
            "username": "webhook-bot",
            "api_key": "bot-key-456",
            "created_by_id": "550e8400-e29b-41d4-a716-446655440000",
            "created_by": "alice",
            "emoji": None,
            "webhook_url": webhook_url,
        },
        status_code=201,
    )

    response = client.create_bot(username="webhook-bot", webhook_url=webhook_url)

    assert response.username == "webhook-bot"
    assert response.webhook_url == webhook_url


def test_get_my_bots(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test getting user's bots."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/bots/me",
        json=[
            {
                "id": "550e8400-e29b-41d4-a716-446655440010",
                "username": "bot1",
                "api_key": "bot-key-1",
                "created_by_id": "550e8400-e29b-41d4-a716-446655440000",
                "created_by": "alice",
                "emoji": "🤖",
                "webhook_url": None,
                "created_at": "2025-10-16T12:00:00Z",
            },
            {
                "id": "550e8400-e29b-41d4-a716-446655440011",
                "username": "bot2",
                "api_key": "bot-key-2",
                "created_by_id": "550e8400-e29b-41d4-a716-446655440000",
                "created_by": "alice",
                "emoji": "🦾",
                "webhook_url": None,
                "created_at": "2025-10-16T12:30:00Z",
            },
        ],
    )

    bots = client.get_my_bots()

    assert len(bots) == 2
    assert bots[0].username == "bot1"
    assert bots[1].username == "bot2"
    assert bots[0].emoji == "🤖"


def test_update_bot(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test updating a bot."""
    client.api_key = "test-key-123"
    bot_id = "550e8400-e29b-41d4-a716-446655440010"
    httpx_mock.add_response(
        method="PATCH",
        url=f"http://test.example.com/bots/{bot_id}",
        json={
            "id": bot_id,
            "username": "bot1",
            "api_key": "bot-key-1",
            "created_by_id": "550e8400-e29b-41d4-a716-446655440000",
            "created_by": "alice",
            "emoji": "🦾",
            "webhook_url": "https://example.com/new/webhook",
            "created_at": "2025-10-16T12:00:00Z",
        },
    )

    response = client.update_bot(
        bot_id=bot_id, emoji="🦾", webhook_url="https://example.com/new/webhook"
    )

    assert response.id == bot_id
    assert response.emoji == "🦾"
    assert response.webhook_url == "https://example.com/new/webhook"


def test_delete_bot(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test deleting a bot."""
    client.api_key = "test-key-123"
    bot_id = "550e8400-e29b-41d4-a716-446655440010"
    httpx_mock.add_response(
        method="DELETE",
        url=f"http://test.example.com/bots/{bot_id}",
        status_code=204,
    )

    # Should not raise an exception
    client.delete_bot(bot_id=bot_id)


def test_regenerate_bot_api_key(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test regenerating bot API key."""
    client.api_key = "test-key-123"
    bot_id = "550e8400-e29b-41d4-a716-446655440010"
    httpx_mock.add_response(
        method="POST",
        url=f"http://test.example.com/bots/{bot_id}/regenerate-api-key",
        json={"message": "API key regenerated", "api_key": "new-bot-key-123"},
    )

    response = client.regenerate_bot_api_key(bot_id=bot_id)

    assert response["message"] == "API key regenerated"
    assert response["api_key"] == "new-bot-key-123"


# Admin Role Management Tests


def test_admin_assign_role(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test admin assigning role to user."""
    client.api_key = "test-key-123"
    user_id = "550e8400-e29b-41d4-a716-446655440000"
    httpx_mock.add_response(
        method="PATCH",
        url=f"http://test.example.com/admin/users/{user_id}/role",
        json={
            "username": "alice",
            "role": "admin",
            "message": "Role assigned successfully",
        },
    )

    response = client.admin_assign_role(user_id=user_id, role="admin")

    assert response.username == "alice"
    assert response.role.value == "admin"
    assert response.message == "Role assigned successfully"


def test_admin_invite_user(httpx_mock: HTTPXMock, client: TokenBowlClient) -> None:
    """Test admin inviting user by email."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/admin/invite",
        json={
            "email": "newuser@example.com",
            "role": "member",
            "message": "Invitation sent successfully",
        },
    )

    response = client.admin_invite_user(
        email="newuser@example.com",
        signup_url="https://app.example.com/signup",
        role="member",
    )

    assert response.email == "newuser@example.com"
    assert response.role.value == "member"
    assert response.message == "Invitation sent successfully"


# User ID vs Username Tests (Regression Prevention)


def test_get_user_profile_with_user_id(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test getting user profile by user_id (not username)."""
    client.api_key = "test-key-123"
    user_id = "550e8400-e29b-41d4-a716-446655440000"
    httpx_mock.add_response(
        method="GET",
        url=f"http://test.example.com/users/{user_id}",
        json={
            "id": user_id,
            "username": "alice",
            "role": "member",
            "logo": "claude.png",
            "bot": False,
            "viewer": False,
        },
    )

    profile = client.get_user_profile(user_id=user_id)

    assert profile.id == user_id
    assert profile.username == "alice"


def test_admin_get_user_with_user_id(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test admin getting user by user_id (not username)."""
    client.api_key = "test-key-123"
    user_id = "550e8400-e29b-41d4-a716-446655440000"
    httpx_mock.add_response(
        method="GET",
        url=f"http://test.example.com/admin/users/{user_id}",
        json={
            "id": user_id,
            "username": "alice",
            "role": "member",
            "email": "alice@example.com",
            "api_key": "user-key-123",
            "webhook_url": None,
            "logo": None,
            "viewer": False,
            "admin": False,
            "bot": False,
            "emoji": None,
            "created_at": "2025-10-16T12:00:00Z",
        },
    )

    profile = client.admin_get_user(user_id=user_id)

    assert profile.id == user_id
    assert profile.username == "alice"
    assert profile.email == "alice@example.com"


def test_admin_update_user_with_user_id(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test admin updating user by user_id (not username)."""
    from token_bowl_chat.models import AdminUpdateUserRequest

    client.api_key = "test-key-123"
    user_id = "550e8400-e29b-41d4-a716-446655440000"
    httpx_mock.add_response(
        method="PATCH",
        url=f"http://test.example.com/admin/users/{user_id}",
        json={
            "id": user_id,
            "username": "alice",
            "role": "member",
            "email": "newemail@example.com",
            "api_key": "user-key-123",
            "webhook_url": None,
            "logo": None,
            "viewer": False,
            "admin": False,
            "bot": False,
            "emoji": "🐰",
            "created_at": "2025-10-16T12:00:00Z",
        },
    )

    update_request = AdminUpdateUserRequest(email="newemail@example.com", emoji="🐰")
    profile = client.admin_update_user(user_id=user_id, update_request=update_request)

    assert profile.id == user_id
    assert profile.email == "newemail@example.com"
    assert profile.emoji == "🐰"


def test_admin_delete_user_with_user_id(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test admin deleting user by user_id (not username)."""
    client.api_key = "test-key-123"
    user_id = "550e8400-e29b-41d4-a716-446655440000"
    httpx_mock.add_response(
        method="DELETE",
        url=f"http://test.example.com/admin/users/{user_id}",
        status_code=204,
    )

    # Should not raise an exception
    client.admin_delete_user(user_id=user_id)


# Message description field test


def test_message_response_has_description_field(
    httpx_mock: HTTPXMock, client: TokenBowlClient
) -> None:
    """Test that MessageResponse includes description field."""
    client.api_key = "test-key-123"
    httpx_mock.add_response(
        method="POST",
        url="http://test.example.com/messages",
        json={
            "id": "msg-1",
            "from_user_id": "550e8400-e29b-41d4-a716-446655440000",
            "from_username": "alice",
            "to_username": None,
            "content": "Hello!",
            "message_type": "room",
            "description": "room message from alice",
            "timestamp": "2025-10-16T12:00:00Z",
        },
        status_code=201,
    )

    response = client.send_message("Hello!")

    assert response.description == "room message from alice"


# Public endpoint test for get_available_logos


def test_get_available_logos_no_auth_required(
    httpx_mock: HTTPXMock,
) -> None:
    """Test that get_available_logos works without authentication."""
    # Create client without API key
    client_no_auth = TokenBowlClient(base_url="http://test.example.com")

    httpx_mock.add_response(
        method="GET",
        url="http://test.example.com/logos",
        json=["claude-color.png", "openai.png"],
    )

    logos = client_no_auth.get_available_logos()

    assert len(logos) == 2
    assert "claude-color.png" in logos



"""Synchronous client for Token Bowl Chat Server."""

import os
from typing import Any, cast

import httpx

from .exceptions import (
    AuthenticationError,
    ConflictError,
    NetworkError,
    NotFoundError,
    RateLimitError,
    ServerError,
    TimeoutError,
    ValidationError,
)
from .models import (
    AdminMessageUpdate,
    AdminUpdateUserRequest,
    AssignRoleRequest,
    AssignRoleResponse,
    BotProfileResponse,
    ConversationResponse,
    CreateBotRequest,
    CreateBotResponse,
    InviteUserRequest,
    InviteUserResponse,
    MessageResponse,
    PaginatedConversationsResponse,
    PaginatedMessagesResponse,
    PublicUserProfile,
    SendMessageRequest,
    StytchAuthenticateRequest,
    StytchAuthenticateResponse,
    StytchLoginRequest,
    StytchLoginResponse,
    UnreadCountResponse,
    UpdateBotRequest,
    UpdateLogoRequest,
    UpdateUsernameRequest,
    UpdateWebhookRequest,
    UserProfileResponse,
    UserRegistration,
    UserRegistrationResponse,
)


class TokenBowlClient:
    """Synchronous client for Token Bowl Chat Server.

    This client provides a Pythonic interface to the Token Bowl Chat Server API
    with full type hints and error handling.

    Example:
        >>> client = TokenBowlClient(api_key="your-api-key")
        >>> client.send_message("Hello, world!")
    """

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str = "https://api.tokenbowl.ai",
        timeout: float = 30.0,
    ) -> None:
        """Initialize the Token Bowl client.

        Args:
            api_key: API key for authentication (optional, defaults to TOKEN_BOWL_CHAT_API_KEY env var)
            base_url: Base URL of the Token Bowl server (optional)
            timeout: Request timeout in seconds
        """
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key or os.getenv("TOKEN_BOWL_CHAT_API_KEY")
        self.timeout = timeout
        self._client = httpx.Client(timeout=timeout)

    def __enter__(self) -> "TokenBowlClient":
        """Context manager entry."""
        return self

    def __exit__(self, *args: object) -> None:
        """Context manager exit."""
        self.close()

    def close(self) -> None:
        """Close the underlying HTTP client."""
        self._client.close()

    def _get_headers(self) -> dict[str, str]:
        """Get request headers including authentication if available.

        Returns:
            Dictionary of HTTP headers

        Raises:
            AuthenticationError: If API key is required but not set
        """
        headers = {"Content-Type": "application/json"}
        if self.api_key:
            headers["X-API-Key"] = self.api_key
        return headers

    def _handle_response(self, response: httpx.Response) -> None:
        """Handle HTTP response errors.

        Args:
            response: HTTP response object

        Raises:
            AuthenticationError: For 401 errors
            NotFoundError: For 404 errors
            ConflictError: For 409 errors
            ValidationError: For 422 errors
            RateLimitError: For 429 errors
            ServerError: For 5xx errors
        """
        if response.status_code < 400:
            return

        try:
            error_data = response.json()
            error_message = str(error_data)
        except Exception:
            error_message = response.text or f"HTTP {response.status_code}"

        if response.status_code == 401:
            raise AuthenticationError(error_message, response)
        elif response.status_code == 404:
            raise NotFoundError(error_message, response)
        elif response.status_code == 409:
            raise ConflictError(error_message, response)
        elif response.status_code == 422:
            raise ValidationError(error_message, response)
        elif response.status_code == 429:
            raise RateLimitError(error_message, response)
        elif response.status_code >= 500:
            raise ServerError(error_message, response)
        else:
            response.raise_for_status()

    def _request(
        self,
        method: str,
        path: str,
        requires_auth: bool = False,
        **kwargs: Any,
    ) -> httpx.Response:
        """Make an HTTP request.

        Args:
            method: HTTP method
            path: API endpoint path
            requires_auth: Whether this endpoint requires authentication
            **kwargs: Additional arguments to pass to httpx

        Returns:
            HTTP response object

        Raises:
            AuthenticationError: If auth is required but API key not set
            NetworkError: For network connectivity issues
            TimeoutError: For request timeouts
        """
        url = f"{self.base_url}{path}"
        headers = self._get_headers()

        if requires_auth and not self.api_key:
            raise AuthenticationError("API key required for this operation")

        try:
            response = self._client.request(method, url, headers=headers, **kwargs)
            self._handle_response(response)
            return response
        except httpx.TimeoutException as e:
            raise TimeoutError(f"Request timed out: {e}") from e
        except httpx.NetworkError as e:
            raise NetworkError(f"Network error: {e}") from e

    def register(
        self,
        username: str,
        webhook_url: str | None = None,
        logo: str | None = None,
    ) -> UserRegistrationResponse:
        """Register a new user and get an API key.

        Args:
            username: Username to register (1-50 characters)
            webhook_url: Optional webhook URL for notifications
            logo: Optional logo filename from available logos

        Returns:
            User registration response with API key

        Raises:
            ConflictError: If username already exists
            ValidationError: If input validation fails
        """
        registration = UserRegistration(
            username=username, webhook_url=webhook_url, logo=logo
        )
        response = self._request(
            "POST",
            "/register",
            json=registration.model_dump(exclude_none=True),
        )
        return UserRegistrationResponse.model_validate(response.json())

    def send_message(
        self,
        content: str,
        to_username: str | None = None,
    ) -> MessageResponse:
        """Send a message to the room or as a direct message.

        Args:
            content: Message content (1-10000 characters)
            to_username: Optional recipient username for direct messages

        Returns:
            Created message response

        Raises:
            AuthenticationError: If not authenticated
            NotFoundError: If recipient doesn't exist
            ValidationError: If input validation fails
        """
        message_request = SendMessageRequest(content=content, to_username=to_username)
        response = self._request(
            "POST",
            "/messages",
            requires_auth=True,
            json=message_request.model_dump(exclude_none=True),
        )
        return MessageResponse.model_validate(response.json())

    def get_messages(
        self,
        limit: int = 50,
        offset: int = 0,
        since: str | None = None,
    ) -> PaginatedMessagesResponse:
        """Get recent room messages with pagination.

        Args:
            limit: Maximum number of messages to return (default: 50)
            offset: Number of messages to skip (default: 0)
            since: ISO timestamp to get messages after

        Returns:
            Paginated list of messages with metadata

        Raises:
            AuthenticationError: If not authenticated
            ValidationError: If parameters are invalid
        """
        params: dict[str, Any] = {"limit": limit, "offset": offset}
        if since is not None:
            params["since"] = since

        response = self._request(
            "GET",
            "/messages",
            requires_auth=True,
            params=params,
        )
        return PaginatedMessagesResponse.model_validate(response.json())

    def get_direct_messages(
        self,
        limit: int = 50,
        offset: int = 0,
        since: str | None = None,
    ) -> PaginatedMessagesResponse:
        """Get direct messages for the current user with pagination.

        Args:
            limit: Maximum number of messages to return (default: 50)
            offset: Number of messages to skip (default: 0)
            since: ISO timestamp to get messages after

        Returns:
            Paginated list of direct messages with metadata

        Raises:
            AuthenticationError: If not authenticated
            ValidationError: If parameters are invalid
        """
        params: dict[str, Any] = {"limit": limit, "offset": offset}
        if since is not None:
            params["since"] = since

        response = self._request(
            "GET",
            "/messages/direct",
            requires_auth=True,
            params=params,
        )
        return PaginatedMessagesResponse.model_validate(response.json())

    def get_users(self) -> list[PublicUserProfile]:
        """Get list of all chat users with their display info.

        Viewer users are excluded as they cannot receive messages.

        Returns:
            List of user profiles with logos, emojis, and bot status

        Raises:
            AuthenticationError: If not authenticated
        """
        response = self._request("GET", "/users", requires_auth=True)
        return [PublicUserProfile.model_validate(user) for user in response.json()]

    def get_online_users(self) -> list[PublicUserProfile]:
        """Get list of users currently connected via WebSocket with their display info.

        Returns:
            List of online user profiles with logos, emojis, and bot status

        Raises:
            AuthenticationError: If not authenticated
        """
        response = self._request("GET", "/users/online", requires_auth=True)
        return [PublicUserProfile.model_validate(user) for user in response.json()]

    def get_available_logos(self) -> list[str]:
        """Get list of available logo filenames.

        This is a public endpoint - no authentication required.

        Returns:
            List of available logo filenames
        """
        response = self._request("GET", "/logos", requires_auth=False)
        return cast(list[str], response.json())

    def update_my_logo(self, logo: str | None = None) -> dict[str, str]:
        """Update the current user's logo.

        Args:
            logo: Logo filename from available logos, or None to clear

        Returns:
            Success message with updated logo

        Raises:
            AuthenticationError: If not authenticated
            ValidationError: If logo is invalid
        """
        logo_request = UpdateLogoRequest(logo=logo)
        response = self._request(
            "PATCH",
            "/users/me/logo",
            requires_auth=True,
            json=logo_request.model_dump(exclude_none=True),
        )
        return cast(dict[str, str], response.json())

    # Stytch Authentication Methods

    def send_magic_link(
        self, email: str, username: str | None = None
    ) -> StytchLoginResponse:
        """Send a magic link to user's email for passwordless authentication.

        If the email is new, a username must be provided to create an account.
        If the email exists, the username field is ignored.

        Args:
            email: Email address to send magic link to
            username: Optional username for new account creation

        Returns:
            Stytch login response with confirmation message

        Raises:
            ValidationError: If input validation fails
            ServerError: If Stytch is not enabled or request fails
        """
        request = StytchLoginRequest(email=email, username=username)
        response = self._request(
            "POST",
            "/auth/magic-link/send",
            json=request.model_dump(exclude_none=True),
        )
        return StytchLoginResponse.model_validate(response.json())

    def authenticate_magic_link(self, token: str) -> StytchAuthenticateResponse:
        """Authenticate a magic link token and return session information.

        If this is a new user (first time authenticating), creates a user account.
        Returns a session token for future requests and an API key.

        Args:
            token: Magic link token from email

        Returns:
            Authentication response with username, session_token, and api_key

        Raises:
            AuthenticationError: If authentication fails
            ValidationError: If token is invalid
        """
        request = StytchAuthenticateRequest(token=token)
        response = self._request(
            "POST",
            "/auth/magic-link/authenticate",
            json=request.model_dump(),
        )
        return StytchAuthenticateResponse.model_validate(response.json())

    # Unread Message Management

    def get_unread_messages(
        self, limit: int = 50, offset: int = 0
    ) -> list[MessageResponse]:
        """Get unread room messages for the current user.

        Args:
            limit: Maximum number of messages to return (default: 50)
            offset: Number of messages to skip (default: 0)

        Returns:
            List of unread room messages

        Raises:
            AuthenticationError: If not authenticated
        """
        params: dict[str, Any] = {"limit": limit, "offset": offset}
        response = self._request(
            "GET",
            "/messages/unread",
            requires_auth=True,
            params=params,
        )
        return [MessageResponse.model_validate(msg) for msg in response.json()]

    def get_unread_direct_messages(
        self, limit: int = 50, offset: int = 0
    ) -> list[MessageResponse]:
        """Get unread direct messages for the current user.

        Args:
            limit: Maximum number of messages to return (default: 50)
            offset: Number of messages to skip (default: 0)

        Returns:
            List of unread direct messages

        Raises:
            AuthenticationError: If not authenticated
        """
        params: dict[str, Any] = {"limit": limit, "offset": offset}
        response = self._request(
            "GET",
            "/messages/direct/unread",
            requires_auth=True,
            params=params,
        )
        return [MessageResponse.model_validate(msg) for msg in response.json()]

    def get_unread_count(self) -> UnreadCountResponse:
        """Get count of unread messages for the current user.

        Returns:
            Unread message counts (room, direct, and total)

        Raises:
            AuthenticationError: If not authenticated
        """
        response = self._request(
            "GET",
            "/messages/unread/count",
            requires_auth=True,
        )
        return UnreadCountResponse.model_validate(response.json())

    def mark_message_read(self, message_id: str) -> None:
        """Mark a message as read.

        Args:
            message_id: ID of the message to mark as read

        Raises:
            AuthenticationError: If not authenticated
            NotFoundError: If message doesn't exist
        """
        self._request(
            "POST",
            f"/messages/{message_id}/read",
            requires_auth=True,
        )

    def mark_all_messages_read(self) -> dict[str, int]:
        """Mark all messages as read for the current user.

        Returns:
            Dictionary with count of messages marked as read

        Raises:
            AuthenticationError: If not authenticated
        """
        response = self._request(
            "POST",
            "/messages/mark-all-read",
            requires_auth=True,
        )
        return cast(dict[str, int], response.json())

    # User Profile Management

    def get_my_profile(self) -> UserProfileResponse:
        """Get the current user's profile information.

        Returns:
            User profile with email, API key, and other information

        Raises:
            AuthenticationError: If not authenticated
        """
        response = self._request(
            "GET",
            "/users/me",
            requires_auth=True,
        )
        return UserProfileResponse.model_validate(response.json())

    def get_user_profile(self, user_id: str) -> PublicUserProfile:
        """Get public profile for a specific user.

        Returns public information (username, logo, emoji, bot, viewer status)
        without sensitive data (API key, email, webhook URL).

        Args:
            user_id: User UUID to retrieve

        Returns:
            Public user profile

        Raises:
            AuthenticationError: If not authenticated
            NotFoundError: If user not found or invalid UUID
        """
        response = self._request(
            "GET",
            f"/users/{user_id}",
            requires_auth=True,
        )
        return PublicUserProfile.model_validate(response.json())

    def update_my_username(self, username: str) -> UserProfileResponse:
        """Update the current user's username.

        Args:
            username: New username

        Returns:
            Updated user profile

        Raises:
            AuthenticationError: If not authenticated
            ConflictError: If username already exists
            ValidationError: If username is invalid
        """
        request = UpdateUsernameRequest(username=username)
        response = self._request(
            "PATCH",
            "/users/me/username",
            requires_auth=True,
            json=request.model_dump(),
        )
        return UserProfileResponse.model_validate(response.json())

    def update_my_webhook(self, webhook_url: str | None) -> dict[str, str]:
        """Update the current user's webhook URL.

        Args:
            webhook_url: New webhook URL, or None to clear

        Returns:
            Success message with updated webhook URL

        Raises:
            AuthenticationError: If not authenticated
            ValidationError: If webhook URL is invalid
        """
        request = UpdateWebhookRequest(webhook_url=webhook_url)
        response = self._request(
            "PATCH",
            "/users/me/webhook",
            requires_auth=True,
            json=request.model_dump(exclude_none=True),
        )
        return cast(dict[str, str], response.json())

    def regenerate_api_key(self) -> dict[str, str]:
        """Regenerate the current user's API key.

        This generates a new API key and invalidates the old one.
        The old API key will no longer work for authentication.

        Returns:
            Success message with new API key

        Raises:
            AuthenticationError: If not authenticated
        """
        response = self._request(
            "POST",
            "/users/me/regenerate-api-key",
            requires_auth=True,
        )
        return cast(dict[str, str], response.json())

    # Admin Methods

    def admin_get_all_users(self) -> list[UserProfileResponse]:
        """Admin: Get all users with full profile information.

        Returns:
            List of all user profiles

        Raises:
            AuthenticationError: If not authenticated or not admin
        """
        response = self._request(
            "GET",
            "/admin/users",
            requires_auth=True,
        )
        return [UserProfileResponse.model_validate(user) for user in response.json()]

    def admin_get_user(self, user_id: str) -> UserProfileResponse:
        """Admin: Get a specific user's full profile.

        Args:
            user_id: User UUID to retrieve

        Returns:
            User profile

        Raises:
            AuthenticationError: If not authenticated or not admin
            NotFoundError: If user not found or invalid UUID
        """
        response = self._request(
            "GET",
            f"/admin/users/{user_id}",
            requires_auth=True,
        )
        return UserProfileResponse.model_validate(response.json())

    def admin_update_user(
        self, user_id: str, update_request: AdminUpdateUserRequest
    ) -> UserProfileResponse:
        """Admin: Update any user's profile fields.

        Args:
            user_id: User UUID to update
            update_request: Fields to update

        Returns:
            Updated user profile

        Raises:
            AuthenticationError: If not authenticated or not admin
            NotFoundError: If user not found or invalid UUID
            ValidationError: If update data is invalid
        """
        response = self._request(
            "PATCH",
            f"/admin/users/{user_id}",
            requires_auth=True,
            json=update_request.model_dump(exclude_none=True),
        )
        return UserProfileResponse.model_validate(response.json())

    def admin_delete_user(self, user_id: str) -> None:
        """Admin: Delete a user.

        Args:
            user_id: User UUID to delete

        Raises:
            AuthenticationError: If not authenticated or not admin
            NotFoundError: If user not found or invalid UUID
        """
        self._request(
            "DELETE",
            f"/admin/users/{user_id}",
            requires_auth=True,
        )

    def admin_get_message(self, message_id: str) -> MessageResponse:
        """Admin: Get a specific message by ID.

        Args:
            message_id: Message ID to retrieve

        Returns:
            Message details

        Raises:
            AuthenticationError: If not authenticated or not admin
            NotFoundError: If message not found
        """
        response = self._request(
            "GET",
            f"/admin/messages/{message_id}",
            requires_auth=True,
        )
        return MessageResponse.model_validate(response.json())

    def admin_update_message(self, message_id: str, content: str) -> MessageResponse:
        """Admin: Update message content.

        Args:
            message_id: Message ID to update
            content: New message content

        Returns:
            Updated message

        Raises:
            AuthenticationError: If not authenticated or not admin
            NotFoundError: If message not found
            ValidationError: If content is invalid
        """
        request = AdminMessageUpdate(content=content)
        response = self._request(
            "PATCH",
            f"/admin/messages/{message_id}",
            requires_auth=True,
            json=request.model_dump(),
        )
        return MessageResponse.model_validate(response.json())

    def admin_delete_message(self, message_id: str) -> None:
        """Admin: Delete a message.

        Args:
            message_id: Message ID to delete

        Raises:
            AuthenticationError: If not authenticated or not admin
            NotFoundError: If message not found
        """
        self._request(
            "DELETE",
            f"/admin/messages/{message_id}",
            requires_auth=True,
        )

    def admin_assign_role(self, user_id: str, role: str) -> AssignRoleResponse:
        """Admin: Assign a role to a user.

        Args:
            user_id: User UUID to assign role to
            role: Role to assign (admin, member, viewer, bot)

        Returns:
            Role assignment response

        Raises:
            AuthenticationError: If not authenticated or not admin with ASSIGN_ROLES permission
            NotFoundError: If user not found or invalid UUID
            ValidationError: If role is invalid
        """
        from .models import Role

        request = AssignRoleRequest(role=Role(role))
        response = self._request(
            "PATCH",
            f"/admin/users/{user_id}/role",
            requires_auth=True,
            json=request.model_dump(),
        )
        return AssignRoleResponse.model_validate(response.json())

    def admin_invite_user(
        self, email: str, signup_url: str, role: str = "member"
    ) -> InviteUserResponse:
        """Admin: Invite a user by email using Stytch magic link.

        Sends a magic link invitation to the specified email address. When the user clicks
        the link and authenticates, they will be automatically registered with the specified role.

        Args:
            email: Email address to invite
            signup_url: URL to redirect to after clicking magic link (e.g., https://app.example.com/signup)
            role: Role to assign to the invited user (default: member)

        Returns:
            Invite response with confirmation

        Raises:
            AuthenticationError: If not authenticated or not admin
            ServerError: If Stytch is not enabled or invitation fails
            ValidationError: If email or role is invalid
        """
        from .models import Role

        request = InviteUserRequest(
            email=email,
            role=Role(role),
            signup_url=signup_url,
        )
        response = self._request(
            "POST",
            "/admin/invite",
            requires_auth=True,
            json=request.model_dump(),
        )
        return InviteUserResponse.model_validate(response.json())

    # Bot Management Methods

    def create_bot(
        self,
        username: str,
        emoji: str | None = None,
        webhook_url: str | None = None,
    ) -> CreateBotResponse:
        """Create a new bot (members and admins only).

        Args:
            username: Bot username (1-50 characters)
            emoji: Optional emoji for the bot
            webhook_url: Optional webhook URL for notifications

        Returns:
            Created bot information with API key

        Raises:
            AuthenticationError: If not authenticated or not member/admin
            ConflictError: If username already exists
            ValidationError: If input validation fails
        """
        request = CreateBotRequest(
            username=username,
            emoji=emoji,
            webhook_url=webhook_url,
        )
        response = self._request(
            "POST",
            "/bots",
            requires_auth=True,
            json=request.model_dump(exclude_none=True),
        )
        return CreateBotResponse.model_validate(response.json())

    def get_my_bots(self) -> list[BotProfileResponse]:
        """Get all bots created by the current user.

        Returns:
            List of bots created by this user

        Raises:
            AuthenticationError: If not authenticated
        """
        response = self._request(
            "GET",
            "/bots/me",
            requires_auth=True,
        )
        return [BotProfileResponse.model_validate(bot) for bot in response.json()]

    def update_bot(
        self,
        bot_id: str,
        emoji: str | None = None,
        webhook_url: str | None = None,
    ) -> BotProfileResponse:
        """Update a bot's configuration (owner or admin only).

        Args:
            bot_id: Bot UUID to update
            emoji: Optional emoji for the bot
            webhook_url: Optional webhook URL for notifications

        Returns:
            Updated bot profile

        Raises:
            AuthenticationError: If not authenticated or don't own the bot
            NotFoundError: If bot not found or invalid UUID
            ValidationError: If update data is invalid
        """
        request = UpdateBotRequest(
            emoji=emoji,
            webhook_url=webhook_url,
        )
        response = self._request(
            "PATCH",
            f"/bots/{bot_id}",
            requires_auth=True,
            json=request.model_dump(exclude_none=True),
        )
        return BotProfileResponse.model_validate(response.json())

    def delete_bot(self, bot_id: str) -> None:
        """Delete a bot (owner or admin only).

        Args:
            bot_id: Bot UUID to delete

        Raises:
            AuthenticationError: If not authenticated or don't own the bot
            NotFoundError: If bot not found or invalid UUID
        """
        self._request(
            "DELETE",
            f"/bots/{bot_id}",
            requires_auth=True,
        )

    def regenerate_bot_api_key(self, bot_id: str) -> dict[str, str]:
        """Regenerate a bot's API key (owner or admin only).

        Args:
            bot_id: Bot UUID

        Returns:
            Success message with new API key

        Raises:
            AuthenticationError: If not authenticated or don't own the bot
            NotFoundError: If bot not found or invalid UUID
        """
        response = self._request(
            "POST",
            f"/bots/{bot_id}/regenerate-api-key",
            requires_auth=True,
        )
        return cast(dict[str, str], response.json())

    # Conversation methods

    def create_conversation(
        self,
        title: str | None = None,
        description: str | None = None,
        message_ids: list[str] | None = None,
    ) -> ConversationResponse:
        """Create a new conversation.

        Args:
            title: Optional conversation title (max 200 chars)
            description: Optional conversation description
            message_ids: List of message IDs to include in conversation

        Returns:
            Created conversation

        Raises:
            ValidationError: If validation fails
            AuthenticationError: If not authenticated
        """
        from .models import CreateConversationRequest

        request_data = CreateConversationRequest(
            title=title,
            description=description,
            message_ids=message_ids or [],
        )
        response = self._request(
            "POST",
            "/conversations",
            json_data=request_data.model_dump(exclude_none=True),
            requires_auth=True,
        )
        from .models import ConversationResponse

        return ConversationResponse.model_validate(response.json())

    def get_conversations(
        self, limit: int = 100, offset: int = 0
    ) -> PaginatedConversationsResponse:
        """Get conversations created by current user (or all if viewer).

        Viewers can see all conversations from all users.
        Regular users can only see their own conversations.

        Args:
            limit: Maximum number of conversations to return (default: 100)
            offset: Number of conversations to skip (default: 0)

        Returns:
            Paginated list of conversations

        Raises:
            AuthenticationError: If not authenticated
        """
        from .models import PaginatedConversationsResponse

        response = self._request(
            "GET",
            "/conversations",
            params={"limit": limit, "offset": offset},
            requires_auth=True,
        )
        return PaginatedConversationsResponse.model_validate(response.json())

    def get_conversation(self, conversation_id: str) -> ConversationResponse:
        """Get a specific conversation.

        Viewers can view any conversation.
        Regular users can only view their own conversations.

        Args:
            conversation_id: Conversation UUID

        Returns:
            Conversation details

        Raises:
            NotFoundError: If conversation not found
            AuthenticationError: If not authenticated or not authorized
        """
        from .models import ConversationResponse

        response = self._request(
            "GET",
            f"/conversations/{conversation_id}",
            requires_auth=True,
        )
        return ConversationResponse.model_validate(response.json())

    def update_conversation(
        self,
        conversation_id: str,
        title: str | None = None,
        description: str | None = None,
        message_ids: list[str] | None = None,
    ) -> ConversationResponse:
        """Update a conversation.

        Args:
            conversation_id: Conversation UUID
            title: Optional new title
            description: Optional new description
            message_ids: Optional new list of message IDs

        Returns:
            Updated conversation

        Raises:
            NotFoundError: If conversation not found
            ValidationError: If validation fails
            AuthenticationError: If not authenticated or not authorized
        """
        from .models import UpdateConversationRequest

        request_data = UpdateConversationRequest(
            title=title,
            description=description,
            message_ids=message_ids,
        )
        response = self._request(
            "PATCH",
            f"/conversations/{conversation_id}",
            json_data=request_data.model_dump(exclude_none=True),
            requires_auth=True,
        )
        from .models import ConversationResponse

        return ConversationResponse.model_validate(response.json())

    def delete_conversation(self, conversation_id: str) -> None:
        """Delete a conversation.

        Args:
            conversation_id: Conversation UUID

        Raises:
            NotFoundError: If conversation not found
            AuthenticationError: If not authenticated or not authorized
        """
        self._request(
            "DELETE",
            f"/conversations/{conversation_id}",
            requires_auth=True,
        )

    def health_check(self) -> dict[str, str]:
        """Check server health status.

        Returns:
            Health status dictionary
        """
        response = self._request("GET", "/health")
        return cast(dict[str, str], response.json())
